
from __future__ import print_function

# ImageD11_v0.4 Software for beamline ID11
# Copyright (C) 2005  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Interface between Tkinter gui and the actual useful code.

There should be no scientific algorithms (eventually) on
the gui side of this class.

This class will eventually offer macro recording capability.
"""

import logging, sys

# Things to offer from gui
from ImageD11 import peakmerge, indexing, transformer, eps_sig_solver

# To autoconvert arrays to lists for Java XMLRPC
RETURN_NUMERICS = False
import numpy as np
TYPE_NUMERIC = type(np.zeros(1))

class guicommand:
    """
    Keeps a log of all commands issued - separates gui code from
    algorithmical code
    """
    def __init__(self):
        self.objects = { "peakmerger" : peakmerge.peakmerger(),
                         "transformer": transformer.transformer(),
                         "indexer"    : indexing.indexer(),
                         "solver"     : eps_sig_solver.solver(), 
                         }

        self.commandscript = \
"""# Create objects to manipulate - they hold your data
#
from ImageD11 import peakmerge, indexing, transformer, eps_sig_solver
mypeakmerger = peakmerge.peakmerger()
mytransformer = transformer.transformer()
myindexer = indexing.indexer()
mysolver  = eps_sig_solver.solver()
#
# Your work starts here:
#
"""

    def execute(self, obj, command, *args, **kwds):
        """
        Pass in object as string [peakmerger|transformer|indexer]
        Pass in command as string, getattr(command) will be used
        Returns the return value of the function....

        TODO : change this interface???
             eg : works - returns True
                          you look for self.lastreturned
                  fails - returns False
                          you look for self.lasttraceback
        """
        if obj not in list(self.objects.keys()):
            raise Exception("ERROR! Unknown command object")
        o = self.objects[obj]
        ran = "my%s.%s("% (obj, command)
        if command.find("."):
            subobjs = command.split(".")[:-1]
            for s in subobjs:
                o = getattr(o,s)
            command = command.split(".")[-1]
        func = getattr(o,command)
        try:
            addedcomma = ""
            for a in args:
                ran="%s %s %s"%(ran,addedcomma,repr(a))
                addedcomma=","
            for k,v in list(kwds.items()):
                ran="%s %s %s=%s "%(ran,addedcomma,k,v)
                addedcomma=","
            ran+=" )\n"
            logging.debug("Running: "+ran)
            sys.stdout.flush()
            ret = func(*args, **kwds)

        except:
            logging.error("Exception occurred " + "self" + str(self) +
                "obj" + str(obj)+ "command" + str(command) +
                "func" + str(func) + "args" + str(args) + "kwds" + str(kwds))
            import traceback
            traceback.print_exc()
            return "Exception occured in the python " + ran
        self.commandscript+=ran
        return ret

    def getdata(self, obj, name):
        """
        Allows access to "live" data in the objects wrapped

        By passing references back you can circumvent the
        cleanliness of the interface. Please dont.

        Returns object.name
        """
        if obj not in list(self.objects.keys()):
            raise Exception("ERROR! Unknown command object")
        attribute = getattr(self.objects[obj],name)
        if RETURN_NUMERICS:
            # Normally python will get this
            logging.debug("python return array")
            return attribute
        if type(attribute) == TYPE_NUMERIC:
            # Java gets this for arrays
            logging.debug("Java return list")
            return attribute.tolist()
        return attribute

    def gethistory(self):
        """
        Returns the history of commands run by the gui commander
        """
        return self.commandscript




from ImageD11 import cImageD11
import numpy as np, time
import unittest


def closest_pyvec( X ):
    nv = X.shape[0]
    nd = X.shape[1]
    res = np.zeros( nv, np.int )
    for i in range(nv):
        dif = X - X[i]
        err = (dif*dif).sum(axis=1)
        err[i] = 1e9
        res[i] = np.argmin( err )
    return res

class test_closest_vec( unittest.TestCase ):
    def setUp( self ):
        np.random.seed( 42 )
        self.data = np.random.random( (2048,12) )

    def test_same_as_python( self ):
        t1 = time.time()
        r1 = closest_pyvec( self.data )
        t2 = time.time()
        r2 = np.zeros( self.data.shape[0], 'i' )
        cImageD11.closest_vec( self.data, r2 )
        t3 = time.time()
        print( "python",t2-t1,"c",t3-t2)
        self.assertTrue( (r1 == r2).all() )

if __name__=="__main__":
    unittest.main()
            





# check for avx2 and import cImageD11 module
import platform, sys, struct
from ImageD11 import cImageD11_sse2
if cImageD11_sse2.got_AVX:  
    from ImageD11 import cImageD11_avx
    print("Using AVX codes")
    from ImageD11.cImageD11_avx import *
else:
    from ImageD11.cImageD11_sse2 import *

# For 32 or 64 bits
nbyte = struct.calcsize("P") # 4 or 8

if nbyte == 8:
    put_incr = put_incr64

if nbyte == 4:
    put_incr = put_incr32

# Add / fix the docstrings

from ImageD11 import cImageD11_docstrings

def fix_doc( oldstring, to_be_added ):
    if oldstring.find("Wrapper for"):
        head, tail = oldstring.split("Wrapper for")
        iname = tail.find("\n")
        return head + tail[:iname] + " " + to_be_added + tail[iname:]
    else:
        return oldstring + to_be_added


for name in cImageD11_docstrings.__all__:
    doc = getattr( cImageD11_docstrings, name )
    if name in globals():
        func = globals()[name]
        func.__doc__ = fix_doc( func.__doc__, doc )
        




from __future__ import print_function

# ImageD11_v0.4 Software for beamline ID11
# Copyright (C) 2015  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import numpy as np
from ImageD11 import grain, transform, cImageD11, indexing, unitcell
import scipy.optimize
import math, time, sys, logging

def unit(a):
    """ Normalise vector """
    return a/np.sqrt(np.dot(a,a))

def get_tth( ds, wvln):
    """ Convert 1/d to Bragg angle """
    return 2 * np.degrees( np.arcsin( ds * wvln / 2 ) )


class indexer:
    """
    A class for searching for orientation matrices
    """
    def __init__(self,
                 transformpars,
                 colfile,
                 ):
        """
        """
        self.transformpars = transformpars
        self.loadcolfile(colfile)        
        self.setcell()
        self.reset()

    def setcell(self):
        """ Sets the unit cell parameters for indexing """
        cp = [self.transformpars.get("cell_%s"%(s)) for s in
              "_a _b _c alpha beta gamma".split()]
        self.unitcell = unitcell.unitcell(cp, 
                        self.transformpars.get("cell_lattice_[P,A,B,C,I,F,R]"))
        

    def loadcolfile(self, colfile):
        self.cf = colfile
        self.updatecolfile()
        for label in "gx gy gz omega tth eta xl yl zl".split():
            assert label in self.cf.titles, label

    def updatecolfile(self):
        if "xl" not in self.cf.titles:
            if "sc" in self.cf.titles:
                pks = self.cf.sc, self.cf.fc
            elif "xc" in self.cf.titles: 
                pks = self.cf.xc, self.cf.yc
            else:
                raise Exception("peaks file misses xc or sc")
            xl,yl,zl = transform.compute_xyz_lab( pks,
                                    **self.transformpars.parameters)
            self.cf.addcolumn(xl,"xl")
            self.cf.addcolumn(yl,"yl")
            self.cf.addcolumn(zl,"zl")
        self.peaks_xyzT = np.array((self.cf.xl,self.cf.yl,self.cf.zl)).T.copy()
        om = self.cf.omega
        osign = self.transformpars.get("omegasign")
        tth, eta = transform.compute_tth_eta_from_xyz(
            self.peaks_xyzT.T, om*osign,
            **self.transformpars.parameters)
        self.cf.addcolumn( tth, "tth", )
        self.cf.addcolumn( eta, "eta", )
        gx, gy, gz = transform.compute_g_vectors(
            tth, eta, om*osign,
            wvln  = self.transformpars.get("wavelength"),
            wedge = self.transformpars.get("wedge"),
            chi   = self.transformpars.get("chi") )
        self.cf.addcolumn(gx, "gx")
        self.cf.addcolumn(gy, "gy")
        self.cf.addcolumn(gz, "gz")           
        self.cf.addcolumn( np.sqrt( gx * gx + 
                                    gy * gy + 
                                    gz * gz ),
                           "modg")
       
    
    def reset(self):
        """ when pars change or colfile changes etc """
        if "ring" not in self.cf.titles:
            self.cf.addcolumn( -np.ones(self.cf.nrows), "ring" )
            self.cf.addcolumn( 42*np.ones(self.cf.nrows), "ringerr" )
        if "labels" not in self.cf.titles:
            self.cf.addcolumn( -np.ones(self.cf.nrows), "labels" )

       

    def tthcalc(self, hkls = None):
        """
        Computes the twotheta for a unit cell given a list of 
        hkls or from the unit cell generating a list of hkls.
        """
        if hkls is None:
            dslimit = self.cf.modg.max()
            wvln = self.transformpars.get( "wavelength" )
            ttol = self.transformpars.get( "fit_tolerance" ) # tth units
            # at limit this ttol make which ds range?
            tthlim = 2 * np.degrees( np.arcsin( dslimit * wvln / 2 ) )
            dstol  = 2 * np.sin( np.radians( (tthlim + ttol) / 2 )
                                 ) / wvln - dslimit
            self.unitcell.makerings(dslimit+dstol/2, tol = dstol)
        else:
            self.unitcell.peaks = [ (self.unitcell.ds(h),h) for h in hkls ]
            self.unitcell.makerings( dslimit, tol=dstol ) # , newpeaks=False )
        self.unitcell.ringtth = [ 2 * np.degrees( np.arcsin( ds * wvln / 2 ) )
                                  for ds
                                  in self.unitcell.ringds ]

    def assigntorings(self):
        """
        Assign the g-vectors to hkl rings that are in self.unitcell
        """
        self.cf.ring[:]=-1
        self.cf.ringerr[:]=42.
        tol = self.transformpars.get( "fit_tolerance" )
        for i,tth in enumerate(self.unitcell.ringtth):
           diff = self.cf.tth - tth
           adiff = np.abs(diff)
           mask = (adiff < tol) & (adiff < self.cf.ringerr)
           if mask.sum()>0:
              self.cf.ring[mask] = i
              self.cf.ringerr[mask] = diff[mask]
        # Report on assignments
        print("Ring     (  h,  k,  l) Mult  total indexed to_index  ")
        # try reverse order instead
        dsr = self.unitcell.ringds
        for j in range(len(dsr))[::-1]:
            ind = self.cf.ring == j
            n_indexed  = (self.cf.labels[ind] >  -0.5).sum()
            n_to_index = (self.cf.labels[ind] <  -0.5).sum()
            h=self.unitcell.ringhkls[dsr[j]][0]
            print("Ring %-3d (%3d,%3d,%3d)  %3d  %5d  %5d  %5d"%(\
                j,h[0],h[1],h[2],len(self.unitcell.ringhkls[dsr[j]]),
                     ind.sum(),n_indexed,n_to_index))
        print("Total peaks",self.cf.nrows,"assigned",(self.cf.ring>=0).sum())


    def pairs(self, hkl1, hkl2, cos_tol = 0.02, hkl_tol = 0.1):
        """
        We only look for reflection pairs matching a single hkl pairing
        """
        import time
        start = time.time()
        w = self.transformpars.get( "wavelength" )
        tth1 = get_tth(self.unitcell.ds(hkl1), w) 
        tth2 = get_tth(self.unitcell.ds(hkl2), w)
        tthtol = self.transformpars.get( "fit_tolerance" )
        allinds = np.arange( self.cf.nrows )
        ind1 = allinds[ abs(self.cf.tth - tth1) < tthtol ]
        ind2 = allinds[ abs(self.cf.tth - tth2) < tthtol ]
        angle, cosangle = self.unitcell.anglehkls( hkl1, hkl2 )
        print("Angle, cosangle",angle,cosangle,hkl1,hkl2)
        assert angle > 1 and angle < 179, "hkls are parallel"
        g = np.array( (self.cf.gx, self.cf.gy, self.cf.gz), np.float )
        n = g/self.cf.modg
        gvf = g.T.copy()
        n1 = n[:,ind1]
        n2 = n[:,ind2]
        pairs = []
        j = np.arange( n2.shape[1] )
        # from unitcell.orient
        h1c=unit(np.dot( self.unitcell.B, hkl1 ))
        h2c=unit(np.dot( self.unitcell.B, hkl2 ))
        t1c=unit(h1c)
        t3c=unit(np.cross(h1c,h2c))
        t2c=unit(np.cross(h1c,t3c))
        T_c =  np.array( [t1c, t2c, t3c] )
        T_g = np.zeros((3,3))
        for i,this_n1 in enumerate(n1.T):
           cosa = np.dot( this_n1, n2 )
           goodones = j[abs(cosa - cosangle) < cos_tol]
           # t1 is along g1
           # t2 is plane of both: g1x(g1xg2)
           # t3 is perpendicular to both
           #if i%100==0:
           #   print i,time.time()-start,len(n1.T)
           for k in goodones:
               this_n2 = n2[:,k]
               T_g[0] = this_n1
               T_g[2] = unit( np.cross( this_n1, this_n2 ) )
               T_g[1] = unit( np.cross( this_n1, T_g[2]) )
               U = np.dot( T_g.T, T_c)
               ub =  np.dot( U, self.unitcell.B)
               ubi = np.linalg.inv( ub )
               ubio = ubi.copy()
               npks = cImageD11.score(ubi,gvf,hkl_tol)
               pairs.append( (ind1[i], ind2[k], U, ubi ) )
               print(npks, end=' ') 
               
               ubi, trans = self.refine( ubi, np.zeros(3,np.float), tol=hkl_tol )
               inds, hkls = self.assign( ubi, trans, hkl_tol )
               ubi, trans = self.refine( ubi, trans, inds = inds, hkls= hkls, tol=hkl_tol )
               print(npks, ubi)
               print("cell: ",6*"%.6f "%(  indexing.ubitocellpars(ubi) ))
               print("position: ",trans)
               print()
        self.pairscache=pairs
        print(time.time()-start,"for",len(pairs),n1.shape, n2.shape)
        return pairs


    def assign(self, ubi, translation, tol):
        gv = np.zeros(self.peaks_xyzT.shape, np.float )
        cImageD11.compute_gv( self.peaks_xyzT,
                    self.cf.omega,
                    self.transformpars.get('omegasign'),
                    self.transformpars.get('wavelength'),
                    self.transformpars.get('wedge'),  
                    self.transformpars.get('chi'),
                    translation,
                    gv)
        hkl = np.dot( ubi, gv.T )
        hkli = np.floor( hkl + 0.5 )
        e = (hkl - hkli)
        e = (e*e).sum(axis=0)
        inds = np.compress( e < tol*tol, np.arange(len(hkl[0])) )
        return inds , hkli[:,inds]


    NC = 0
    def gof( self, p, *args ):
        self.NC += 1
        try:
            hkls, inds, peaks_xyzT, gobs, omega = args
        except:
            print(args, len(args))
            raise
        p.shape = 4,3
        ub = p[:3]
        t = p[3]
        gcalc = np.dot( ub, hkls )
        cImageD11.compute_gv( peaks_xyzT,
                              omega,
                              self.transformpars.get('omegasign'),
                              self.transformpars.get('wavelength'),
                              self.transformpars.get('wedge'),  
                              self.transformpars.get('chi'),
                              t,
                              gobs)
        #print gobs
        #print gcalc
        #print (gobs-gcalc).ravel()
        #1/0
        e = (gcalc - gobs.T).ravel()
        p.shape = 12,
 #       print p-0.1,(e*e).sum()
        return e#(e*e).sum()
        
    def refine(self, ubi, translation=(0,0,0) ,
               inds = None, hkls = None,
               tol = 0.05):
        """
        Fit ubi and translation
        ihkls = array of peak_index, h, k, l
        tol = hkl tolerance to use in fitting
        """
        import time
        self.NC =0
        start = time.time()
        if inds is None:
            inds , hkls = self.assign(  ubi, translation, tol )
        ub = np.linalg.inv( ubi )
        x0 = np.array( list( ub.ravel() ) + list(translation ) )
        fun = self.gof
        args = (hkls, inds, self.peaks_xyzT[inds],
                np.zeros(self.peaks_xyzT[inds].shape ),
                self.cf.omega[inds])
        def Dfun( x0, *args ):
            epsilon = np.ones(12)*1e-6
            epsilon[-3:] = 1.
            return deriv( x0, fun, epsilon, *args)
        res, ier = scipy.optimize.leastsq( fun, x0, args, Dfun,
                                           col_deriv=True)
        ub = np.reshape(res[:9], (3,3))
        t = res[-3:]
        ubi = np.linalg.inv( ub )
        return ubi, t
 
def deriv( xk, f, epsilon, *args):
    f0 = f(*((xk,) + args))
    grad = np.zeros((len(xk),len(f0)), float)
    ei = np.zeros((len(xk),), float)
    for k in range(len(xk)):
        ei[k] = 1.0
        d = epsilon * ei
        grad[k] = (f(*((xk + d,) + args)) - f0) / d[k]
        ei[k] = 0.0
    return grad

if __name__=="__main__":
    from ImageD11.columnfile import columnfile
    from ImageD11.parameters import read_par_file
    import ImageD11.grain
    import sys
    p = read_par_file(sys.argv[1])
    c = columnfile( sys.argv[2] )
    i = indexer( p, c )
    if sys.argv[3][:3] == "fit":
        #                                            0                   1                  2           3    4             5
        # \test\simul_1000_grains>python ..\..\ImageD11\indexer.py Al1000\Al1000.par Al1000\Al1000.flt fit allgrid.map allgridfitscipy.map
        gl = ImageD11.grain.read_grain_file( sys.argv[4] )
        inds = np.arange( len(gl), dtype=np.int )
        allhkls = np.array( (c.h, c.k, c.l) )
        for k,g in enumerate(gl):
            if len(sys.argv[3]) == len("fit"):
                for j, tol in enumerate([0.05,0.02,0.01,0.0075]):
                    inds, hkls = i.assign( g.ubi, g.translation, tol )
                    ubi, t = i.refine(   g.ubi, 
                                         translation=g.translation,
                                         inds = inds, hkls = hkls,
                                         tol = tol)  
                    g.translation = t
                    g.set_ubi( ubi )
            else:
                inds = np.compress( c.labels == k , inds )
                hkls = np.compress( c.labels == k , allhkls )
                for j in range(3):
                    ubi, t = i.refine(   g.ubi, 
                                         translation=g.translation,
                                         inds = inds, hkls = hkls,
                                         tol = tol)  
                    g.translation = t
                    g.set_ubi( ubi )
            print(k, len(inds),6*"%.8f "%(indexing.ubitocellpars(ubi)))
            print("\t",t)
        ImageD11.grain.write_grain_file( sys.argv[5], gl )
    else:
        i.updatecolfile()
        i.tthcalc()
        print("Calling assign")
        i.assigntorings()
        hkl1 = [int(h) for h in sys.argv[3].split(",")]
        hkl2 = [int(h) for h in sys.argv[4].split(",")]
        print("Calling pairs")
        i.pairs(hkl1, hkl2)


   # e.g. python indexer.py ../test/demo/eu3.pars ../test/demo/eu.flt 2,0,0 0,0,4




import numpy as np
from ImageD11 import indexing, fft_index_refac

o = indexing.indexer()
o.readgvfile("test.vecs")
g = fft_index_refac.grid( 128, 1., 5.)
g.gv_to_grid_new( o.gv )
g.fft()
g.props()
g.peaksearch( "pairFFT.pks" )
g.read_peaks( "pairFFT.pks" )
tol=0.1
tol2=tol*tol

def refine_vector( v, gv, tol=0.25, ncycles=25, precision=1e-6 ):
    """
    Input
    v  :  real space lattice vector
    gv : scattering vectors (no errors)
    tol : tolerance for initial assignment to peak
    ncycles : how much to refine
    precision : when to stop refining in v += s when |s|/|v| < precision
    Returns
    vref : refined real space lattice vector

    Uses:
    hr = v[0]*gv[i,0] + v[0]*gv[i,1] + v[1]*gv[i,2]  1 per peak
    hi = round(hr) = calc                            1 per peak
    diff = calc - hr                                 1 per peak
    gradient = d(diff)/dv[i] = gv[j,i]               3 per peak
    matrix = gradient[i].gradient[j]                 3x3 peaks^2
    rhs = gradient . diff                            3

    Solve for shifts and iterates reducing tolerance as 1/(ncycles+1)
    Peaks that change index during refinement are removed
    Stops on : 
       no peaks reassigned
       ncycles runs out
       |shift| / |vec| < precision
    """
    assert gv.shape[1] == 3
    vref = v.copy()
    mg = (gv*gv).sum(axis=1)
    mold = None
    wt = 1./mg # + mg.max())
    for i in range(ncycles):
        hr = np.dot( vref, gv.T )   # npks
        hi = np.round( hr )         # npks
        diff = hi - hr              # npks
        m = abs(diff) < tol/(i+1)   #  keep or not ?
        diff = (diff*wt)[m]
        grad = (gv.T*wt).T[m]
        rhs = np.dot( grad.T, diff )
        mat = np.dot( grad.T, grad )
        U,s,V = np.linalg.svd( mat )
        one_over_s = np.where( s/s.max() < 1e-6, 1, 1./s )
        mati = np.dot( U, np.dot(np.diag( one_over_s ), V ) )
        vshft = np.dot( mati, rhs )
        vref = vref + vshft
        if mold is not None:
            changed = mold ^ m   # XOR
            nch = changed.sum()
            # Down weight peaks that are different to before for next round
            if nch == 0:
                break
            wt[ changed ] = 0.
            #print nch,i
        mold = m
        slen = np.sqrt((vshft * vshft).sum())
        vlen = np.sqrt((vref*vref).sum())
#        print "shift = ",slen,vlen,slen/vlen, tol/(i+1)
        if abs(slen/vlen) < precision:
            break
    return vref
    
def scr(v): 
    h=np.dot(v, o.gv.T)
    return h-np.round(h)

    

idx = []
drl = []
vs = []
for v in g.UBIALL:
    v = refine_vector( v, o.gv )
    vs.append(v)
    hr = np.dot( v, o.gv.T )
    hi = np.round( hr )
    dr = (hr - hi)
    drlv2 = dr*dr
    drl.append( drlv2 )
    m = drlv2 < tol2
    idx.append( m )
    print( v, m.sum() )

idx = np.array( idx ) # dims vecs x npk
pair = np.zeros( (len(idx),len(idx)))
for i in range(len(g.UBIALL)):
    for j in range(i, len(g.UBIALL)):
        pair[i,j] = pair[j,i] = (idx[i]*idx[j]).sum()
        

import pylab as pl
pl.imshow( pair, interpolation='nearest' )
pl.show()





from __future__ import print_function
# testing for the peaksearch command line script

# These are not really unit tests

import numpy as np

def w(name,image,omega=0):

    f=open(name,"wb")
    h = "{%1021s}\n"%("""Omega = %.2f ;\n Dim_1 = 256 ;\n Dim_2 = 256 ;
    DataType = UnsignedShort ;
    ByteOrder = LowByteFirst ;
    Image = 1 ;


    \n\n"""%(omega))
    f.write(h.encode('utf-8'))
    f.write(np.ravel(image).astype(np.uint16).tostring())
    f.close()

def makedata():
    image=np.ones((256,256),np.uint16)*20
    image[20:53,56:99]=20000 # big spot
    w("data0000.edf",image,0)
    image[20:53,56:99]=20 # put back to 20
    image[20:30,56:66]=10000 # two disconnected
    image[43:53,89:99]=10000
    w("data0001.edf",image,1)
    image[20:53,56:99]=20000 # put back to 20000
    w("data0002.edf",image,2)
    image[20:53,56:99]=20 # put back to 20
    w("data0003.edf",image,3)
    w("dark_good.edf",image*0+1)
    w("dark_bad.edf",image*0+65534)

def losedata():
    for i in range(4):
        name = "data%04d.edf"%(i)
        print("removing",name)
        os.remove(name)
    os.remove("dark_bad.edf")
    os.remove("dark_good.edf")

# FIXME - fit this into unit tests and kill the combinatorial explosion
#       - add checks to see if the answers are OK or not.

import os, sys
ps = os.path.join("..","scripts")
ps = os.path.join(ps,"peaksearch.py")

print (ps)

cmds = [sys.executable + " " + ps + " -n data -f 0 "]


cmds = [cmd + "-l %d "%(last) for last in [0,1,3] for cmd in cmds]


thresholds = ["0","5500"]

cmds = [cmd + "-t %s "%(t) for t in thresholds for cmd in cmds]

cmds = cmds + [ cmd + "-t 1500 -t 25000 " for cmd in cmds]

# cmds = cmds + [cmd + "-F edf " for cmd in cmds]

cmds = cmds + [cmd + "-d %s "%(d) for d in ["dark_good.edf","dark_bad.edf"] for cmd in cmds]

# floods as dark
cmds = cmds + [cmd + "-O %s "%(d) for d in ["dark_good.edf","dark_bad.edf"] for cmd in cmds]

# one with spline , one perfect

clist = [ cmd + "-p Y " for cmd in cmds] + [cmd + "-s spatial2k.spline" for cmd in cmds]

import os

def testcmdlines(clist):
    makedata()
    print(len(clist))
    for c in clist:
        print (c)
        ret = os.system(c)
        if ret != 0:
            raise Exception("Bombed")
    losedata()

if __name__=="__main__":
    # takes ages
    # FIXME
    import sys
    if len(sys.argv)>1 and sys.argv[1] == "ALL":
        testcmdlines(clist)

    testcmdlines([ sys.executable + ' ' + ps + ' -n data -f 0 -l 3 -t 15000 -t 1000 -p Y -t 0'])
    assert len(open("peaks_t0.flt").readlines())==2
    assert len(open("peaks_t1000.flt").readlines())==2
    assert len(open("peaks_t15000.flt").readlines())==3





from matplotlib.pylab import *
from matplotlib.colors import cnames
keys = cnames.keys()




def HSVtoRGB(  h, s, v ):
    """ http://www.cs.rit.edu/~ncs/color/t_convert.html """
    if s == 0 :
    	return v,v,v
    h /= 60;		#	 sector 0 to 5
    i = floor( h );
    f = h - i;		#	 factorial part of h
    p = v * ( 1 - s );
    q = v * ( 1 - s * f );
    t = v * ( 1 - s * ( 1 - f ) );

    if i == 0: return v,t,p	
    if i == 1: return q,v,p
    if i == 2: return p,v,t		
    if i == 3: return p,q,v
    if i ==4 : return t,p,v
    return v,p,q	








z = []
i = []
g=0
filloff = 0.
for line in open(sys.argv[1],"r").readlines():
    if line.find("ubi[")>0 and len(z)>0:
        a = argsort
        g+=12.
        c = HSVtoRGB(g%360,0.8,0.9)
        # (g*0.4)%1, (g*0.45)%1, (g*0.55)%1
        if max(c)>1.:
            print c,g
        if z[0]>filloff and max(i)>1.5e9:        
            cf = HSVtoRGB(g%360,0.3,0.95)

            fill([0]+i+[0],array([z[0]]+z+[z[-1]])-3.15,label="%d"%(g),facecolor=cf,
                    edgecolor=c)
            filloff = z[-1]
        plot(i,array(z)-3.15,marker="+",linestyle="-", color = c)
        z=[]
        i=[]
    if line.find("ubi[")>0:
        z=[]
        i=[]
    if line.find("520\\scan")==-1:
        continue
    items = line.split()
    if int(items[0])<15:
        continue
    zh = float(items[1])
    if len(z) > 0 and zh > z[-1]:
        z.append(float(items[1]))
        diff = zh - z[0]
        i.append(float(items[-1]))
    else:
        if len(z) == 0:
            z.append(float(items[1]))
            diff = zh - z[0]
            i.append(float(items[-1]))
        else:
            print line.rstrip(),z,zh,len(z)
#legend(loc=0)
title("Diffracted intensity versus height")
xlabel("Total intensity")
ylabel("Height in sample")
show()





"""
Determine Bravias lattice for UBI

Return transformation matrix


"""


from ImageD11.unitcell import unitcell
from ImageD11.indexing import ubitocellpars
from numpy import array


def fi_triclinic(ubi):
    """ Ensure we have the 3 shortest non-coplanar vectors """
    return ubi

def fi_monoclinic(ubi, cell, mult):
    """ 
    Choose vectors having mult == 2
    These are either along b or in ac plane
    If centred then one mult 4 vector is shorter and cell par
    """
    possibles = []
    for i in range(len(mult)):
        if mult[i] == 2:
            possibles.append( cell.ringhkls(cell.ringds[i])[0] )
    possibles = [cell.ring]
    

def ferraris_and_ivaldi(ubi, tol = 0.001):
    """
    Acta Cryst (1983) A39 595-596

    Determine the multiplicities of vectors in a computed powder pattern
    Match these to multiplicities according to lattice types
    ... appeals to powder person
    Tolerance is for grouping vector lengths.
    """
    cpar = ubitocellpars(ubi)
    cell = unitcell( cp )
    dlim = cell.ds( array( [7,7,7] ))
    u.makerings( dlim , tol)
    mult = [ len(cell.ringhkls[d]) for d in cell.ringds ]
    mx = max(mult)
    mn = min(mult)
    cases = {
        (2,  2) : fi_triclinic,
        (4,  2) : fi_monoclinic,
        (8,  2) : fi_orthorhombic,
        (12, 2) : fi_rhombohedral,
        (16, 2) : fi_tetragonal,
        (24, 2) : fi_hexagonal,
        (48, 2) : fi_cubic }
        

def katayama(ubi):
    """
    J. Appl. Cryst (1986) 19 69-72
    
    Take four vectors a,b,c and d=-(a+b+c)
    Compute the 6 pairwise dot products (ab, ac, ad, bc, bd, cd)
    Apply transformations (up to 12) and recompute
    
    etc
    """
    pass





# Create objects to manipulate - they hold your data
#
from ImageD11 import indexing
import unittest, sys

class fake():
    def write(self,*a,**k):
        pass
    def flush(self,*a,**k):
        pass

class testKen(unittest.TestCase):
    def test12(self):
        ss=sys.stdout
        sys.stdout = fake()
        try:
            myindexer = indexing.indexer()
            myindexer.readgvfile( 'simAl.gve' )
            myindexer.updateparameters( )
            myindexer.parameterobj.set_parameters(
                {'minpks': '80',
                 'ring_1': '0',
                 'ring_2': '1',
                } )
            myindexer.loadpars( )
            myindexer.assigntorings( )
            myindexer.find( )
            myindexer.scorethem( )
            # myindexer.saveubis(  'test.ubi' )
            myindexer.assigntorings( )
            # myindexer.saveindexing(  'test.idx' )
            self.assertEqual( len( myindexer.ubis ) , 20 )
        finally:
            sys.stdout=ss

if __name__=="__main__":
    unittest.main()




import logging
logging.basicConfig(level=logging.INFO)
# Create objects to manipulate - they hold your data
#
from ImageD11 import peakmerge, indexing, transformer
mypeakmerger = peakmerge.peakmerger()
mytransformer = transformer.transformer()
myindexer = indexing.indexer()
#
# Your work starts here:
#    
mypeakmerger.readpeaks(  'eu.pks' )
mypeakmerger.harvestpeaks(  numlim=(0.0, 1000.0)  , omlim=(-100.0, 100.0)  )
mypeakmerger.mergepeaks( )
mypeakmerger.filter( )
mypeakmerger.savepeaks(  'eu.flt' )
mytransformer.loadfiltered(  'eu.flt' )
mytransformer.loadfileparameters(  'eu2.pars' )
mytransformer.compute_tth_eta( )
mytransformer.addcellpeaks( )
mytransformer.fit( 0.0 , 14.0 )
mytransformer.saveparameters(  'eu3fitted.pars' )
mytransformer.computegv( )
mytransformer.savegv(  'eu3.gve' )
myindexer.readgvfile(  'eu3.gve' )
myindexer.updateparameters( )
p = {'ds_tol': '0.005', 'minpks': '100', 'uniqueness': '0.5', 'hkl_tol': '0.05', 'eta_range': '0.0', 'ring_1': '1', 'wavelength': '0.153684', 'ring_2': '1', 'cosine_tol': '0.002'} 
myindexer.parameterobj.set_parameters(  p )
myindexer.loadpars( )
myindexer.assigntorings( )
myindexer.find( )
myindexer.scorethem( )
myindexer.histogram_drlv_fit( )
myindexer.saveubis(  'eu3.ubi' )
myindexer.saveindexing(  'eu3.idx' )





from __future__ import print_function, division

import time, sys
import h5py, scipy.sparse, numpy as np, pylab as pl
from ImageD11 import cImageD11


# see also sandbox/harvest_pixels.py

NAMES = {
    "filename" : "original filename used to create a sparse frame",
    "intensity" : "corrected pixel values",
    "nlabel": "Number of unique labels for an image labelling",
    "threshold" : "Cut off used for thresholding",
    }


class Container( object ):
    """ More-or-less a fabio image style wrapper
        data = numpy array
        header = metadata to write to h5py attrs """
    def __init__(self, data, **kw ):
        self.data = data
        self.header = kw
        

class sparse_frame( object ):
    """
    Indices / shape mapping
    """
    def __init__(self, row, col, shape, itype=np.uint16, pixels=None):
        """ row = slow direction
        col = fast direction
        shape = size of full image
        itype = the integer type to store the indices
                our c codes currently use unsigned short...
        nnz is implicit as len(row)==len(col)
        pixels = named Containers (arrays + metadata)
        """
        self.check( row, col, shape, itype )
        self.shape = shape
        self.row = np.asarray(row, dtype = itype )
        self.col = np.asarray(col, dtype = itype )
        self.nnz = len(self.row)
        # Things we could have using those indices:
        #   raw pixel intensities
        #   corrected intensities
        #   smoothed pixel intensities
        #   labelling via different algorithms
        self.pixels = {}
        if pixels is not None:
            for name, val in pixels.items():
                # print(val.data.shape, self.nnz)
                assert len(val.data) == self.nnz
                assert isinstance( val, Container )
                self.pixels[name] = val
        
    def check(self, row, col, shape, itype):
        """ Ensure the index data makes sense and fits """
        lo = np.iinfo(itype).min
        hi = np.iinfo(itype).max
        assert len(shape) == 2
        assert shape[0] >= lo and shape[0] < hi
        assert shape[1] >= lo and shape[1] < hi
        assert np.min(row) >= lo and np.max(row) < hi
        assert np.min(col) >= lo and np.max(col) < hi
        assert len(row) == len(col)
        
    def is_sorted(self):
        """ Tests whether the data are sorted into slow/fast order
        rows are slow direction
        columns are fast """
        # TODO: non uint16 cases
        assert self.row.dtype == np.uint16 and \
            cImageD11.sparse_is_sorted( self.row, self.col ) == 0

    def to_dense(self, data, out=None):
        """ returns the full 2D image 
        Does not handle repeated indices
        e.g.  obj.to_dense( obj.pixels['raw_intensity'] )
        """
        if out is None:
            out = np.zeros( self.shape, data.dtype )
        else:
            assert out.shape == self.shape
        assert len(data) == self.nnz
        adr = self.row.astype(np.intp) * self.shape[1] + self.col
        out.flat[adr] = data    
        return out

    def mask( self, msk ):
        """ returns a subset of itself """
        return sparse_frame( self.row[msk],
                             self.col[msk],
                             self.shape, self.row.dtype,
                             pixels= {
                                 name: Container( pixval.data[msk],
                                                  **pixval.header )
                                 for (name, pixval) in self.pixels.items()} )

    def set_pixels( self, name, values ):
        """ Named arrays sharing these labels """
        assert isinstance( values, Container )
        assert len(values.data) == self.nnz
        self.pixels[name] = values

    def sort_by( self, name ):
        """ Not sure when you would do this. For sorting 
        by a peak labelling to get pixels per peak """
        assert name in self.pixels
        order = np.argsort( self.pixels[name].data )
        self.reorder( self, order )

    def sort( self ):
        """ Puts you into slow / fast looping order """
        order = np.lexsort( ( self.col, self.row ) )
        self.reorder( self, order )

    def reorder( self, order ):
        """ Put the pixels into a different order """
        assert len(order) == self.nnz
        self.row = self.row[order]
        self.col = self.col[order]
        self.pixels = { (name, Container( pixval.data[order],
                                          **pixval.header) )
                         for (name, pixval) in self.pixels.items() }
        
    def threshold(self, threshold, name='intensity'):
        """
        returns a new sparse frame with pixels > threshold
        """
        return self.mask( self.pixels.data[name] > threshold )

    def to_hdf_group( frame, group ):
        """ Save a 2D sparse frame to a hdf group """
        itype = np.dtype( frame.row.dtype )
        meta = { "itype"  : itype.name,
                 "shape0" : frame.shape[0],
                 "shape1" : frame.shape[1] }
        for name, value in meta.items():
            group.attrs[name] = value
        #opts = { "compression": "lzf" }
        opts = {}
        group.require_dataset( "row", shape=(frame.nnz,),
                               dtype=itype, **opts )
        group.require_dataset( "col", shape=(frame.nnz,),
                               dtype=itype, **opts )
        group['row'][:] = frame.row
        group['col'][:] = frame.col
        for pxname in frame.pixels:
            group.require_dataset( pxname, shape=(frame.nnz,),
                                   dtype=frame.pixels[pxname].data.dtype,
                                   **opts ) 
            group[pxname][:] = frame.pixels[pxname].data
            for name, value in frame.pixels[pxname].header.items():
                group[pxname].attrs[ name ] = value


    
def from_data_mask( mask, data, header ):
    """
    Create a sparse from a dense array
    """
    assert mask.shape == data.shape
    # using uint16 here - perhaps make this general in the future
    # ... but not for now
    assert data.shape[0] < pow(2,16)-1
    assert data.shape[1] < pow(2,16)-1
    nnz = (mask>0).sum()
    tmp = np.empty( data.shape[0],'i') # tmp hold px per row cumsums
    row = np.empty( nnz, np.uint16 )
    col = np.empty( nnz, np.uint16 )
    cImageD11.mask_to_coo( mask, row, col, tmp )
    intensity = data[ mask > 0 ]
    return sparse_frame( row, col, data.shape, itype=np.uint16,
                  pixels = { "intensity" : Container( intensity, **header ) } )

def from_hdf_group( group ):
    itype = np.dtype( group.attrs['itype'] )
    shape = group.attrs['shape0'], group.attrs['shape1']
    row = group['row'][:] # read it
    col = group['col'][:]
    pixels = {}
    for pxname in list(group):
        if pxname in ["row", "col"]:
            continue
        data = group[pxname][:]
        header = dict( group[pxname].attrs )
        pixels[pxname] = Container( data, **header )
    return sparse_frame( row, col, shape, itype=itype, pixels=pixels )

def sparse_moments( frame, intensity_name, labels_name ):
    """ We rely on a labelling array carrying nlabel metadata (==labels.data.max())"""
    nl = frame.pixels[ labels_name ].header[ "nlabel" ]
    return cImageD11.sparse_blob2Dproperties(
        frame.pixels[intensity_name].data,
        frame.row,
        frame.col,
        frame.pixels[labels_name].data,
        nl )


def overlaps(frame1, labels1, frame2, labels2):
    """
    figures out which label of self matches which label of other
    Returns sparse array of:
    label in self (row)
    label in other (col)
    number of shared pixels (data)
    """
    ki = np.empty( frame1.nnz, 'i' )
    kj = np.empty( frame2.nnz, 'i' )
    npx = cImageD11.sparse_overlaps( frame1.row, frame1.col, ki,
                                     frame2.row, frame2.col, kj)
    # self.data and other.data filled during init
    row = frame1.pixels[labels1].data[ ki[:npx] ]  # my labels
    col = frame2.pixels[labels2].data[ kj[:npx] ]  # your labels
    ect = np.empty( npx, 'i')    # ect = counts of overlaps
    tj  = np.empty( npx, 'i')    # tj = temporary  for sorting
    n1  = frame1.pixels[labels1].header[ "nlabel" ]
    n2  = frame2.pixels[labels2].header[ "nlabel" ]
    tmp = np.empty( max(n1, n2)+1, 'i') # for histogram
    nedge = cImageD11.compress_duplicates( row, col, ect, tj, tmp )
    # overwrites row/col in place
    crow = row[:nedge]
    ccol = col[:nedge]
    cdata = ect[:nedge]
    cedges = scipy.sparse.coo_matrix( ( cdata, (crow, ccol)) )
    return cedges
 
    
def sparse_connected_pixels( frame,
                             label_name="connectedpixels",
                             data_name="intensity",
                             threshold=None ):
    """
    frame = a sparse frame
    label_name = the array to save labels to in that frame
    data_name = an array in that frame
    threshold = float value or take data.threshold
    """
    labels = np.zeros( frame.nnz, "i" )
    if threshold is None:
        threshold = frame.pixels[data_name].header["threshold"]
    nlabel = cImageD11.sparse_connectedpixels(
        frame.pixels[data_name], frame.row, rame.col,
        threshold,  labels )
    px = Container( labels, **{ "data_name": data_name,
                                "threshold": threshold,
                                "nlabel"   : nlabel } )
    frame.set_pixels( label_name, px )


def sparse_localmax( frame,
                     label_name="localmax",
                     data_name = "intensity" ):
    labels = np.zeros( frame.nnz, "i" )
    vmx = np.zeros( frame.nnz, np.float32 )
    imx = np.zeros( frame.nnz, 'i')
    nlabel = cImageD11.sparse_localmaxlabel(
        frame.pixels[data_name].data, frame.row, frame.col,
        vmx, imx, labels )
    px = Container( labels, **{ "data_name" : data_name,
                                "nlabel" : nlabel } )
    frame.set_pixels( label_name, px )

    
                              
        



        





from __future__ import print_function


from ImageD11.columnfile import columnfile
from ImageD11 import transform, parameters, unitcell, grain, gv_general
import numpy as np, time
from matplotlib import pylab


def k_to_g(k, omega, wedge, chi ):
    """ Rotate k vector to g vector """
    #print "k_to_g",k,omega,wedge,chi
    om = np.radians( omega )
    g = np.zeros( k.shape )
    cw = gv_general.chiwedge( chi, wedge )
    cwk = np.dot( cw, k )
    co = np.cos( np.radians( omega )) # 1D
    so = np.sin( np.radians( omega )) # 1D
    g[0,:] =  co*cwk[0,:] + so*cwk[1,:]
    g[1,:] = -so*cwk[0,:] + co*cwk[1,:]
    g[2,:] = cwk[2,:]
    return g

def g_to_k(g, omega, wedge, chi):
    """ Rotate g vector to k vector """
    co = np.cos(np.radians(omega))
    so = np.sin(np.radians(omega))
    k = np.zeros( (3,len(omega)))
    k[0,:] = co * g[0,:] - so * g[1,:]
    k[1,:] = so * g[0,:] + co * g[1,:]
    k[2,:] = g[2,:]
    cw = gv_general.chiwedge( chi, wedge )
    k = np.dot( cw, k )
    return k


def grain_x0( omega, wedge, chi, tx, ty, tz ):
    """ Grain origin positions at Omega rotation """
    co = np.cos(np.radians(omega))
    so = np.sin(np.radians(omega))
    x0 = np.zeros(( 3, omega.shape[0]) )
    x0[0] = co*tx - so*ty
    x0[1] = so*tx + co*ty
    x0[2] = tz
    cw = gv_general.chiwedge( chi, wedge )
    x0 = np.dot( cw.T, x0 )
    return x0


def XL_to_gv_old( omega, wedge, chi, XL, wavelength, tx, ty, tz):
    """ Historic way to compute gv from peak position """
    beam = np.array( [ 1/wavelength, 0, 0] )
    x0 = grain_x0( omega, wedge, chi, tx, ty, tz)
    s1 = XL - x0
    mods = np.sqrt( (s1*s1).sum(axis=0) ) * wavelength
    k = ((s1/mods).T - beam).T
    g = k_to_g( k, omega, wedge, chi )
    return g

def XL_to_gv( omega, wedge, chi, XL, wavelength, tx, ty, tz):
    """ Compute gv by rotation of experiment around crystal """
    beam = np.zeros( (3,len(omega)), np.float)
    beam[0,:] = 1.0
    XB = k_to_g( beam, omega, wedge, chi)
    #print "XB",XB
    XS = k_to_g( XL, omega, wedge, chi )
    #print "XS",XS
    np.subtract( XS[0], tx, XS[0] )
    np.subtract( XS[1], ty, XS[1] )
    np.subtract( XS[2], tz, XS[2] )
    modxs = np.sqrt( (XS*XS).sum(axis=0))
    g = (XS/modxs - XB)/wavelength
    return g

def test_XL_to_gv():
    """ Check the old and new way give the samy gv """
    omega = np.array([0,20,60,127,243])
    wedge = 0
    chi = 0
    XL = np.ones((15))
    XL.shape=3,5
    wavelength = 1.2345
    T = np.array( [0,0,0] )
    gnew = XL_to_gv( omega, wedge, chi, XL, wavelength, T[0],T[1],T[2])
    gold = XL_to_gv_old( omega, wedge, chi, XL, wavelength, T[0],T[1],T[2])
    print(gnew-gold)

#test_XL_to_gv()


def XSXB_to_gv( XS, XB, tx, ty, tz, wavelength):
    """ 
    XS - position of spot on detector
    XB - direction of X-ray beam
    tx, ty, tz - grain position
    wavelength
    Computes g-vector
    """
    xst = (XS.T - [tx,ty,tz]).T
    modxst = np.sqrt( (xst*xst).sum(axis=0) )
    g = (xst/modxst - XB)/wavelength
    return g

# derivatives:
def d_XSXB_to_gv( XS, XB, tx, ty, tz, wavelength):
    """
    As XSXB_to_gv but with derivative on tx, ty, tz
    """
    xst = XS[0] - tx, XS[1] - ty, XS[2] - tz
    # dxst_dt = -1
    modxst2 = xst[0]*xst[0] + xst[1]*xst[1] + xst[2]*xst[2]
    # dmodxst2 = 2
    modxst = np.sqrt( modxst2 )
    # dmodxst_dt = dmodxst_dxst . dxst_dt
    # dmodxst2_dt = -2*xst[0] , -2*xst[1], -2*xst[2]
    # d(sqrt(f))/dx = 1/2 1/sqrt(f) . df/dx
    dmodxst_dt = -xst[0]/modxst, -xst[1]/modxst , -xst[2]/modxst
    # From XSXB_to_gv
    #    g = np.array((    (xst[0]/modxst - XB[0])/wavelength,\
    #                      (xst[1]/modxst - XB[1])/wavelength,\
    #                      (xst[2]/modxst - XB[2])/wavelength ) )
    g = (xst / modxst - XB ) /wavelength
    #
    # g[0] = (xst[0]/modxst - XB[0])/wavelength
    # g[1] = (xst[1]/modxst - XB[1])/wavelength
    # g[2] = (xst[2]/modxst - XB[2])/wavelength
    #
    # Want dg/dt
    #
    # dg/dt = (dxst/dt*1/modxst + xst.dmodxst_dt/modxst2)/w
    dg0_dtx = -(1./modxst + xst[0]*dmodxst_dt[0]/modxst2)/wavelength
    dg1_dtx = -(xst[0]*dmodxst_dt[1]/modxst2)/wavelength
    dg2_dtx = -(xst[0]*dmodxst_dt[2]/modxst2)/wavelength
    dg0_dty = -(xst[1]*dmodxst_dt[0]/modxst2)/wavelength
    dg1_dty = -(1./modxst + xst[1]*dmodxst_dt[1]/modxst2)/wavelength
    dg2_dty = -(xst[1]*dmodxst_dt[2]/modxst2)/wavelength
    dg0_dtz = -(xst[2]*dmodxst_dt[0]/modxst2)/wavelength
    dg1_dtz = -(xst[2]*dmodxst_dt[1] /modxst2)/wavelength
    dg2_dtz = -(1./modxst + xst[2]*dmodxst_dt[2]/modxst2)/wavelength
    #
           #   1       2       3       4       5       6       7       8       9
    return g,dg0_dtx,dg1_dtx,dg2_dtx,dg0_dty,dg1_dty,dg2_dty,dg0_dtz,dg1_dtz,dg2_dtz


class fittrans(object):

    def __init__(self, fname, parfile):
        self.colfile = columnfile( fname )
        self.pars = parameters.read_par_file( parfile )
        
        self.ds_tol = 0.005
        self.bins=np.arange(0,0.5,0.005)
        print("Setting up")
        self.compute_XLYLZL()
        self.computegv()

    def computegv(self):
        sign = float(self.pars.get('omegasign'))
        tx  = float(self.pars.get('t_x'))
        ty  = float(self.pars.get('t_y'))
        tz  = float(self.pars.get('t_z'))
        wvln= float(self.pars.get('wavelength'))
        self.gv = XSXB_to_gv( self.XS, self.XB, tx, ty, tz, wvln)

    def refine(self, gr, tol=0.05):
        tx,ty,tz = gr.translation
        wvln = float(self.pars.get("wavelength"))
        if hasattr(gr, "pks") and gr.pks is not None:
            #print "Got pks"
            pks = gr.pks
            XS = self.XS[:,pks]
            XB = self.XB[:,pks]
        else:
            #print "New pks"
            XS = self.XS
            XB = self.XB
            pks = np.arange( len(XS[0]) ) 
        ret = d_XSXB_to_gv( XS, XB, tx, ty, tz, wvln)
        gv = ret[0]
        dg0_dt = ret[1],ret[4],ret[7]
        dg1_dt = ret[2],ret[5],ret[8]
        dg2_dt = ret[3],ret[6],ret[9]
        hklr = np.dot( gr.ubi, gv )
        hkli = np.round(hklr)
        hkle = hklr - hkli
        scor = np.sqrt((hkle*hkle).sum(axis=0))
        #print "before",len(pks),pks
        if tol is not None:
            use = np.compress( scor < tol,  np.arange(len(gv[0]) ))
            #print "after",len(pks),pks
            gr.pks = pks[use]
        else:
            use = np.arange( len(gr.pks), dtype=np.int )
        #print "score = ", scor[pks].sum()/len(pks), len(pks), tol
        # peaks to use are those with scor OK
        #
        #  UB.h = gvcalc
        #  dg/dUB = h
        #  dg/dT  = found above
        gcalc = np.dot( gr.ub,  hkli )
        diff = np.take(gv - gcalc, use, axis=1)
        # print diff.shape, pks.shape
        # gv[0],[1],[2] = 3
        # tx, ty, ty    = 3
        # UB00 ... UB22 = 9
        # want derivative of diff w.r.t each variable
        grads = np.zeros( (12, 3, len(use)) )
        for i in range(3):
            for j in range(3): # tx, ty, tz
                #print 1+j*3+i,
                #print ret[1+j*3+i].shape
                grads[j, i] = ret[1+j*3+i][use]
           #     print grads[j,i]
            for j in range(3):
                # gx = 0h + 1k + 2l
                # gy = 3h + 4k + 5l
                # gz = 6h + 7k + 8l
                # i is gx, gy, gz
                # j is ub elements
                grads[3+j+i*3 , i] = hkli[j][use]
           #     print grads[3+j+i*3,i]
        # grains = 12, xyz, pks
        mtx = np.zeros( (12,12) )
        for i in range(12):
            for j in range(i,12):
                for k in range(3):
                    mtx[i,j] += (grads[i,k,:] * grads[j,k,:]).sum()
                if j!=i:
                    mtx[j,i] = mtx[i,j]

        #    mtx = np.dot( grads, grads.T) # vector, outer, ?
        rhs = np.zeros( 12 )
        for i in range(12):
            for k in range(3):
                rhs[i] += (grads[i,k]*diff[k]).sum()
        #print mtx
        # print rhs
        imt = np.linalg.inv( mtx )
        shifts = np.dot( imt, rhs )
        tx = tx - shifts[0]
        ty = ty - shifts[1]
        tz = tz - shifts[2]
        gr.translation = [tx,ty,tz]
        ub = gr.ub.ravel()
        np.add(ub, shifts[3:], ub)
        gr.set_ubi( np.linalg.inv( np.reshape(ub,(3,3))))
        gr.npks = len(use)
        #1/0
        return gr


    def compute_XLYLZL(self):
        self.peaks_xyz = transform.compute_xyz_lab( [self.colfile.sc,
                                                    self.colfile.fc],
                                                  **self.pars.parameters)
        # These are the peak positions in the laboratory frame
        #
        # Now get the beam and peaks in the crystal frame
        wedge = float(self.pars.get('wedge'))
        chi   = float(self.pars.get('chi'))
        omega = self.colfile.omega
        beam = np.zeros( (3,len(omega)))
        beam[0,:] = 1
        self.XB = k_to_g( beam, omega, wedge, chi)
        self.XS = k_to_g( self.peaks_xyz, omega, wedge, chi )



    def makecell(self):
        p = self.pars.parameters
        c =[ float(v) for v in [
                p['cell__a'],
                p['cell__b'],
                p['cell__c'],
                p['cell_alpha'],
                p['cell_beta'],
                p['cell_gamma'] ]]
        self.cell = unitcell.unitcell(c, p['cell_lattice_[P,A,B,C,I,F,R]'])
        self.cell.makerings( limit = self.ds.max(), tol = self.ds_tol )

    def ringassign(self, ds=None):
        if ds is None:
            ds = self.colfile.ds
        diffs = np.ones( self.colfile.nrows, np.float32 )
        ra    = np.ones( self.colfile.nrows, np.int ) - 2
        for i in range(len(self.cell.ringds)):
            testdiff = abs( ds - self.cell.ringds[i] )
            msk = testdiff < diffs
            diffs = np.where( msk, testdiff, diffs )
            ra = np.where( msk, i, ra )
        self.ra = ra
        self.diffs = diffs
        pylab.figure()
        vol = np.sqrt(np.linalg.det(self.cell.g))
        rscp = pow(vol, 1.0/3.0)
        pylab.hist( diffs*rscp, bins = np.arange(0,0.5,0.005) )
        print((diffs*rscp < 0.05).sum())
        return ra, diffs

    def find_triplets(self, peak):
        gx = self.colfile.gx - self.colfile.gx[peak]
        gy = self.colfile.gy - self.colfile.gy[peak]
        gz = self.colfile.gz - self.colfile.gz[peak]
        ds = np.sqrt( gx*gx + gy*gy + gz*gz )
        ra, diffs = self.ringassign( ds )


if __name__=="__main__":
    import sys
    from ImageD11.indexing import ubitocellpars
    o = fittrans( sys.argv[1], sys.argv[2] )
    gl = grain.read_grain_file( sys.argv[3] )
    gref = gl[0]
    import time
    start = time.time()
    gfl = []
    ng = 0
    # Take old peak assignments:
    if 1:
        print("Using existing peak assignments")
        inds = np.arange(o.colfile.nrows,dtype=np.int)
        for i,gref in enumerate(gl):
            gref.pks = np.compress( o.colfile.labels == i, inds )
        tols = [None,]*3
    else:
        tols = [0.05,0.01,0.0075]
    for gref in gl:
        for ii,tol in enumerate(tols):
            #print gref.translation
            gref = o.refine(gref, tol=tol)
            #print ii, gref.translation, gref.npks,
            #print i,gref.npks
#        gref.pks = None
        # re-assign after convergence
#        gref = o.refine( gref, tol=0.0075) 
        gfl.append( gref )
        ng += 1
        print(ng,ng*100.0/len(gl), "%.2f   "%(time.time()-start), gref.npks, end=' ')
        print((3*"%.4f ")%tuple(gref.translation), end=' ')
        print((6*"%.6f ")%ubitocellpars(gref.ubi))
    print(time.time()-start)
    grain.write_grain_file(sys.argv[4], gfl)
#    1/0
#    for i in range(2):
#        o.find_triplets( i*17 )
#    pylab.show()

# ~/ImageD11/test/simul_1000_grains % python ../../sandbox/fittrans.py Al1000/Al1000.flt Al1000/Al1000.par allgrid.map allgridfittrans.map




from __future__ import division, print_function
import numpy as np, scipy.ndimage, sys, fabio
from ImageD11 import cImageD11, labelimage

def gauss_1d( N, center, fwhm ):
    """
    Normalised 1D gaussian peak
    """
    sigma = fwhm / 2.35482
    dx = (np.arange(0, N, 1, dtype=np.float32 ) - center)/sigma
    f = np.exp( -dx*dx / 2 ) / sigma / np.sqrt( 2 * np.pi )
    return f

def lorentz_1d( N, center, fwhm ):
    """
    Normalised 1D Lorentz peak
    """
    dx = (np.arange(0, N, 1, dtype=np.float32 ) - center)
    top = fwhm / 2. / np.pi
    bottom = dx*dx + fwhm*fwhm / 4.
    f = top / bottom
    return f




def add_gaussian_2d( img, center, sigma, weight ):
    """ adds a 2d Gaussian into a test image 
    (missing sig_ij for now)
    """
    ns, nf = img.shape
    i, j = np.mgrid[0:ns, 0:nf]
    di = i.astype(np.float32) - center[0]
    dj = j.astype(np.float32) - center[1]
    s2_2 = 2*sigma*sigma
    arg = (di*di+dj*dj)/s2_2
    np.add( img,
            np.exp( -arg )/s2_2/np.pi,
            img )
    return img
    
# gaussian_2d = np.outer( gauss_1d, gauss_1d )
#
# Separation will be interesting for 2theta, azimuth representations...
#

def make_test_image( dims  = (256,256),
                     sigmas = [3.5,],
                     centers = [(128,128),],
                     weights = [1,] ):
    """ produce a test image with some peaks in it """
    imt = np.zeros( dims, np.float32 )
    for (i,j),sigma,weight in zip(centers, sigmas, weights):
        imt = add_gaussian_2d( imt, (i,j), sigma, weight )
    return imt



class convolutepeak(object):
    """
    Gets ready to convolve your image with some functions
       pk - peak width 
       bg - flat
       peak - function to produce 1d peak 
    Uses these convolutions to do a pixel by pixel peak fit (of a local box)
     ... e.g. attempts to strip the background
    """
    def __init__(self, dims, fwhm, peak=gauss_1d ):
        """
        Fits y = pk + bkg    pixel by pixel...
        """
        # Dimension of pk and bg images to convolve
        N = int(3*fwhm+1)*2+1
        p = peak( N, N//2, fwhm )
        o = np.ones(N, np.float32 )
        # 2D peak
        pk = np.outer( p, p )
        bg = np.outer( o, o )
        # Matrix for a least squares problem fitting peak in this box
        m = [ [ (bg*bg).sum(), (bg*pk).sum()],
              [ (bg*pk).sum(), (pk*pk).sum()] ]
        detM = 1./(m[0][0]*m[1][1] - m[0][1]*m[1][0])
        invM = [[ detM*m[1][1],  -detM*m[0][1]],
                [ -detM*m[1][0], detM*m[0][0]]]
        # convolve me to fit peak
        pkc = invM[1][0]*bg + invM[1][1]*pk
        # ...    pk = np.outer( pk1d, pk1d )
        # ...    bg = np.outer( ones, ones )
        # p0 = invM[1][0] * np.ones( N, np.float32 ) + \
        #      invM[1][1] * gauss_1d( N, N//2, sigma )
        #    print( invM )
        #    return p0
        self.p, self.o, self.invM = p, o, invM
        
    def conv( self, im ):
        """ Fits pk + bg """
        p, o, invM = self.p, self.o, self.invM
        # 2D as a pair of 1D
        # pk
        imp = scipy.ndimage.correlate1d( im, invM[1][1] * p,
                                         axis=0, mode= 'nearest' )
        imp = scipy.ndimage.correlate1d( imp, p,
                                         axis=1, mode= 'nearest' )
        # bg
        imb = scipy.ndimage.correlate1d( im, invM[1][0] * o,
                                         axis=0, mode= 'nearest' )
        imb = scipy.ndimage.correlate1d( imb, o,
                                         axis=1, mode= 'nearest' )
        # fitted
        return imb + imp


def process_edf( fname, bkg , sigma_threshold, sigma_peak, sptfile ):
    im = fabio.open( fname ) 
    cpk = convolutepeak( im.data.shape, sigma_peak )
    lio = labelimage.labelimage( im.data.shape, sptfile )
    # for i in range(im.nframes):
    i = 0
    while True:
        try:
            im = im.next()
        except:
            break
        data = im.data.astype( np.float32 )
        np.subtract( data, bkg, data )
        pkdata = cpk.conv( data )
        s0 = pkdata.std()*sigma_threshold
        m0 = abs(pkdata) < s0
        s1 = pkdata[m0].std()*sigma_threshold
        m1 = abs(pkdata) < s1
        s2 = pkdata[m1].std()*sigma_threshold
        lio.peaksearch( pkdata, s2, i )
        print( "%s[%d] %d"%( im.filename, i, lio.npk ) )
        lio.mergelast()
        i+=1
    lio.finalise()
        
        
    
    
if __name__ == "__main__":
    edf      = sys.argv[1]
    bkgname  = sys.argv[2]
    bkg      =  fabio.open( bkgname ).data
    sptfile  = edf.replace(".edf",".flt")
    try:
        sigma_threshold = float(sys.argv[3])
    except:
        sigma_threshold = 5.
    sigma_peak = 2.
    try:
        fwhm_peak = float(sys.argv[4])
    except:
        fwhm_peak = 3.
    process_edf( edf, bkg, sigma_threshold , fwhm_peak , sptfile)

    






from ImageD11.columnfile import columnfile 
import sys
import numpy as np
import silx.gui.qt , silx.gui.plot3d.SceneWindow
from silx.gui.plot3d.tools.PositionInfoWidget import PositionInfoWidget
from silx.gui.widgets.BoxLayoutDockWidget import BoxLayoutDockWidget




if __name__ == "__main__":
    colf = columnfile( sys.argv[1] )
    colf.parameters.loadparameters( sys.argv[2] )
    colf.updateGeometry()

    qapp = silx.gui.qt.QApplication([])

    # Create a SceneWindow widget
    window = silx.gui.plot3d.SceneWindow.SceneWindow()

    sceneWidget = window.getSceneWidget()
    sceneWidget.setBackgroundColor((0.1, 0.18, 0.08, 1.))
    sceneWidget.setForegroundColor((1., 1., 1., 1.))
    sceneWidget.setTextColor((0.5, 0.7, 0.7, 1.))
    sceneWidget.setProjection('orthographic')

    positionInfo = PositionInfoWidget()
    positionInfo.setSceneWidget(sceneWidget)
    dock = BoxLayoutDockWidget()
    dock.setWindowTitle("Selection Info")
    dock.setWidget(positionInfo)
    window.addDockWidget(silx.gui.qt.Qt.BottomDockWidgetArea, dock)

    x,y,z,values = colf.gx, colf.gy, colf.gz, colf.avg_intensity

    scatter3d = silx.gui.plot3d.items.Scatter3D()
    scatter3d.setData(x, y, z, values)
    # Set scatter3d properties
    scatter3d.getColormap().setName('viridis')  # Use 'magma' colormap
    scatter3d.setSymbol('.') # Use point markers
    scatter3d.setSymbolSize(11)  # Set the size of the markers

    # Add scatter3d to the scene
    sceneWidget.addItem( scatter3d )

    # Set scatter3d transform
    SIZE=1
    scatter3d.setScale(SIZE, SIZE, SIZE)

    window.show()
    # Avoid segfault on exceptions
    sys.excepthook = silx.gui.qt.exceptionHandler
    qapp.exec_()


blob_moments = """fills in the reduced moments in results array.
... FIXME - this would be clearer in python, fast anyway.
"""
bloboverlaps = """determines the overlaps between labels1 and labels2
for an image series. Peaks in labels2 may be merged if they were
joined by a peak on labels1. Results in results1 are accumulated
into results2 if peaks are overlapped.
"""
blobproperties = """fills the array results with properties of each labelled
object described by data (pixel values) and labels. The omega value
is the angle for this frame.
results are FIXME
"""
compress_duplicates = """removes duplicate i,j labels. On entry then
i and j are set as the labels from two images. They are sorted
and on exit i,j hold the unique pairs and oi holds the count
for the number of overlaps. oj and tmp are temporaries.
"""
connectedpixels = """Determines which pixels in data are above the
user supplied threshold and assigns them into connected objects 
which are output in labels. Connectivity is 3x3 box (8) by default
and reduces to a +(4) is con8==0
"""
localmaxlabel = """assigns a label for each pixel so they are grouped
to the local maximum. Equal values choose to assign towards the earlier
value in memory.
cpu arg (1)0=C, (1)1=SSE2, (1)2=AVX2; if > 9 prints timing
"""
sparse_blob2Dproperties = """fills the array results with properties of each labelled
object described by v and labels (pixel values and blob) and positions i,j in the image.
results are: 
  results[ipk,s2D_1]   = sum (1), number of pixels
  results[ipk,s2D_I]   = sum (I), total intensity
  results[ipk,s2D_fI]  = sum (f*I), intensity weighted fast index
  results[ipk,s2D_sI]  = sum (s*I), intensity weighted slow index
  results[ipk,s2D_ffI] = sum (f*f*I), intensity weighted fast^2 index
  results[ipk,s2D_sfI] = sum (s*f*I), intensity weighted slow*fast index
  results[ipk,s2D_ssI] = sum (s*s*I), intensity weighted slow^2 index
"""
sparse_localmaxlabel = """assigns labels to sparse array in sorted coo format
supplied in (v,(i,j)). MV and iMV are temporaries.
single threaded 
"""
sparse_overlaps = """identifies the pixels in i1,j1 which overlap i2,j2.
The list of overlaps is returned in k1/k2 such that i1[k1]==i2[k2]
and j1[k1]==j2[k2]. Probably assumes that sparse_is_sorted was true.
"""
__all__ = ["blob_moments","bloboverlaps","blobproperties","compress_duplicates","connectedpixels","localmaxlabel","sparse_blob2Dproperties","sparse_localmaxlabel","sparse_overlaps"]



from __future__ import print_function, division


# ImageD11_v0.4 Software for beamline ID11
# Copyright (C) 2005  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import numpy as np
from . import cImageD11, unitcell
from xfab.tools import ubi_to_u, u_to_rod, ubi_to_rod

import math, time, sys, logging


def ubi_fit_2pks( ubi, g1, g2):
    """
    Refine a ubi matrix so it matches the pair of g-vectors supplied
    (almost) accounts for cell parameters not being quite right
    """
    ub = np.linalg.inv( ubi )
    h1 = np.round( np.dot( ubi, g1 ) )
    h2 = np.round( np.dot( ubi, g2 ) )
    g3 = np.cross( g1, g2 )
    h3 = np.dot( ubi, g3 )    # do not round to integer
    g1c = np.dot( ub, h1 )
    g2c = np.dot( ub, h2 )
    g3c = np.dot( ub, h3 )
    R = np.outer( g1, h1 ) + np.outer( g2, h2 ) + np.outer( g3, h3 )
    H = np.outer( h1, h1 ) + np.outer( h2, h2 ) + np.outer( h3, h3 )
    ubfit = np.dot( R, np.linalg.inv( H ) )
    ubifit = np.linalg.inv( ubfit )
    return ubifit
   
   
   

def myhistogram(data, bins):
   """
   The numpy histogram api was changed
   So here is an api that will not change
   It is based on that from the old Numeric manual
   """
   n = np.searchsorted( np.sort(data), bins )
   n = np.concatenate( [ n, [len(data)]] )
   return n[1:] - n[:-1]


def readubis(ubifile):
    """read ubifile and return a list of ubi arrays """
    f = open(ubifile, "r")
    ubisread = []
    u = []
    for line in f:
        if line[0]=="#":
            continue
        vals = [ float(x) for x in line.split() ]
        if len(vals) == 3:
            u = u + [vals]
        if len(u)==3:
            ubisread.append(np.array(u))
            u = []
    f.close()
    return ubisread

def write_ubi_file(filename, ubilist):
    """ save 3x3 matrices into file """
    f=open(filename,"w")
    for u in ubilist:
        f.write("%f %f %f\n"  %(u[0][0],u[0][1],u[0][2]))
        f.write("%f %f %f\n"  %(u[1][0],u[1][1],u[1][2]))
        f.write("%f %f %f\n\n"%(u[2][0],u[2][1],u[2][2]))
    f.close()

def ubitocellpars(ubi):
    """ convert ubi matrix to unit cell """
    g = np.dot(ubi, np.transpose(ubi))
    from math import acos, degrees, sqrt
    a = sqrt(g[0,0])
    b = sqrt(g[1,1])
    c = sqrt(g[2,2])
    alpha = degrees( acos(g[1,2]/b/c))
    beta  = degrees( acos(g[0,2]/a/c))
    gamma = degrees( acos(g[0,1]/a/b))
    return a, b, c, alpha, beta, gamma

def ubitoU(ubi):
    """
    convert ubi to Grainspotter style U
    The convention is B as being triangular, hopefully as Busing and Levy
    TODO - make some testcases please!!
    """
    #return np.transpose(np.dot(ubitoB(ubi),ubi))
    return ubi_to_u(ubi)

def ubitoRod(ubi):
    """
    TODO Testcases!!!
    """
#     u = ubitoU(ubi)
#     w, v = np.linalg.eig(u)
#     print 'Eigenvalues'
#     print w
#     print 'Eigen vectors'
#     print v
#     #ehat = v[:,0]
#     #angle = -1*math.acos(np.clip(w[order[1]].real,-1,1))
#     order = np.argsort(w.real)
#     print order
#     ehat = v[:, order[-1]]
#     if order.tolist() != range(3):
#         print 'HHFH'
#         angle = -1*np.arccos(w[order[1]].real)
#     else:
#         angle = np.arccos(w[order[1]].real)
#     Rod = ehat * math.tan(angle/2)
#     return Rod.real
    return ubi_to_rod(ubi)

def ubitoB(ubi):
    """ give the B matrix from ubi """
    g = np.dot(ubi, np.transpose(ubi))
    return np.transpose(np.linalg.inv(np.linalg.cholesky(g)))


def mod_360(theta, target):
    """
    Find multiple of 360 to add to theta to be closest to target
    """
    diff = theta - target
    while diff < -180:
        theta = theta + 360
        diff = theta - target
    while diff > 180:
        theta = theta - 360
        diff = theta - target
    return theta

def calc_drlv2(UBI, gv):
    """
    Get the difference from being integer hkls
    UBI: ubi matrix
    gv: list of g-vectors
    returns drlv2 = (h_calc - h_int)^2
    """
    h = np.dot(UBI, np.transpose(gv))
    hint = np.floor(h + 0.5).astype(np.int) # rounds down
    diff = h - hint
    drlv2 = np.sum(diff * diff,0)
    return drlv2



def refine(UBI, gv, tol, quiet=True):
    """
    Refine an orientation matrix and rescore it.

    From Paciorek et al Acta A55 543 (1999)
       UB = R H-1
    where:
       R = sum_n r_n h_n^t
       H = sum_n h_n h_n^t
       r = g-vectors
       h = hkl indices
    """
    #      print "Orientation and unit cell refinement of"
    #      print "UBI\n",UBI
    #      print "Scores before",self.score(UBI)
    # Need to find hkl indices for all of the peaks which are indexed
    h = np.dot(UBI, np.transpose(gv))
    hint = np.floor(h+0.5).astype(np.int) # rounds down
    diff = h - hint
    drlv2 = np.sum( diff * diff, 0)
    tol = float(tol)
    tol = tol * tol
    # Only use peaks which are assigned to rings for refinement
    ind = np.compress( np.less(drlv2,tol) , np.arange(gv.shape[0]) )
    # scoreb4=ind.shape[0]
    contribs = drlv2[ind]
    try:
        fitb4=math.sqrt(np.sum(contribs)/contribs.shape[0])
        if not quiet:
            logging.debug("Fit before refinement %.8f %5d"% (
                                             fitb4, contribs.shape[0]))
    except:
        logging.error("No contributing reflections for \n%s"%(str(UBI)))
        raise
    # drlv2_old=drlv2
    R=np.zeros((3,3),np.float)
    H=np.zeros((3,3),np.float)
    for i in ind:
        r = gv[i,:]
        k = hint[:,i].astype(np.float)
        #           print r,k
        R = R + np.outer(r,k)
        H = H + np.outer(k,k)
    try:
        HI=np.linalg.inv(H)
        UBoptimal=np.dot(R,HI)
        UBIo=np.linalg.inv(UBoptimal)
    except:
        # A singular matrix - this sucks.
        UBIo=UBI
    h=np.dot(UBIo,np.transpose(gv))
    hint=np.floor(h+0.5).astype(np.int) # rounds down
    diff=h-hint
    drlv2=np.sum(diff*diff,0)
    ind = np.compress( np.less(drlv2,tol), np.arange(gv.shape[0]) )
    # scorelastrefined=ind.shape[0]
    contribs = drlv2[ind]
    try:
        fitlastrefined=math.sqrt(np.sum(contribs)/contribs.shape[0])
        if not quiet:
            logging.debug("after %.8f %5d"%(fitlastrefined,contribs.shape[0]))
    except:
        logging.error("\n\n\n")
        logging.error("No contributing reflections for %s\n",str(UBI))
        logging.error("After refinement, it was OK before ???")
        logging.error("\n\n\n")
        return UBI
        # raise
    #      for i in ind:
    #         print "( %-6.4f %-6.4f %-6.4f ) %12.8f %12.8f"%(\
    #            h[0,i],h[1,i],h[2,i],sqrt(drlv2[i]),sqrt(drlv2_old[i]))
    #      print UBIo
    #      print "Scores after", self.score(UBIo,self.hkl_tol)
    #      print "diff\n",UBI-UBIo
    #      print "Mean drlv now",sum(sqrt(drlv2))/drlv2.shape[0],
    #      print "Mean drlv old",sum(sqrt(drlv2_old))/drlv2_old.shape[0]
    return UBIo

def indexer_from_colfile( colfile, **kwds ):
    uc = unitcell.unitcell_from_parameters( colfile.parameters )
    w = float( colfile.parameters.get("wavelength") )
    gv = np.array( (colfile.gx,colfile.gy,colfile.gz), np.float)
    kwds.update( {"unitcell": uc, "wavelength":w, "gv":gv.T } )
    return indexer( **kwds )


class indexer:
    """
    A class for searching for orientation matrices
    """
    def __init__(self,
            unitcell=None,
            gv=None,
            cosine_tol = 0.002,
            minpks = 10 ,
            hkl_tol=0.01,
            ring_1=1,
            ring_2=2,
            ds_tol=0.005,
            wavelength=-1,
            uniqueness=0.5,
            eta_range=0.,
            max_grains=100):
        """
        Unitcell would be a unitcell object for generating hkls peaks
        gv would be a 3*n array of points in reciprocal space
        The rest of the arguments are parameters.
        """
        # This stop variable allows computation to be run in a thread...
        self.stop = False
        self.unitcell=unitcell
        self.gv=gv
        if gv is not None: # do init
            logging.info('gv: %s %s %s', str(gv), str(gv.shape), str(gv.dtype))
            assert gv.shape[1] == 3
            self.gv = gv.astype( np.float )
            self.ds = np.sqrt( (gv*gv).sum(axis=1) )
            self.ga = np.zeros(len(self.ds),np.int32)-1 # Grain assignments
            self.gvflat=np.ascontiguousarray(gv, np.float)
            self.wedge=0.0 # Default

        self.cosine_tol=cosine_tol
        self.wavelength=wavelength
        self.hkl_tol=hkl_tol
        self.ring_1=ring_1
        self.ring_2=ring_2
        self.uniqueness=uniqueness
        self.minpks=minpks
        self.ds_tol=ds_tol
        self.max_grains=max_grains
        self.eta_range = eta_range
        self.ubis=[]
        self.scores=[]


        # it would make more sense to inherit the parameter object - will
        # have to think about this some more - how general is it?
        from ImageD11 import parameters
        self.parameterobj = parameters.parameters(cosine_tol=self.cosine_tol,
              hkl_tol=self.hkl_tol, ring_1=self.ring_1, ring_2=self.ring_2,
              minpks=self.minpks, uniqueness=self.uniqueness, ds_tol=self.ds_tol,
              wavelength=self.wavelength, eta_range=self.eta_range,
              max_grains=self.max_grains)
        
        # Add a resetting functionality, adapted from
        # stackoverflow.com/questions/4866587/pythonic-way-to-reset-an-objects-variables
        import copy
        self.__pristine_dict = copy.deepcopy( self.__dict__ )

    def reset( self ):
        import copy
        self.__dict__ = copy.deepcopy( self.__pristine_dict )
        

    def loadpars(self,filename=None):
        if filename is not None:
            self.parameterobj.loadparameters(filename)
        self.parameterobj.update_other(self)

    def updateparameters(self):
        self.savepars()
        self.pars=self.parameterobj.parameters

    def savepars(self,filename=None):
        self.parameterobj.update_yourself(self)
        if filename is not None:
            self.parameterobj.saveparameters(filename)

    def out_of_eta_range(self,eta):
        """ decide if an eta is going to be kept """
        e = mod_360(float(eta), 0)
        if e < abs(self.eta_range) and e > -abs(self.eta_range):
            return True
        if e < -180.+abs(self.eta_range) or e > 180.-abs(self.eta_range):
            return True
        return False

    def assigntorings(self):
        """
        Assign the g-vectors to hkl rings
        """
        # rings are in self.unitcell
        limit = np.amax( self.ds )
        print("Assign to rings, maximum d-spacing considered",limit)
        self.unitcell.makerings(limit, tol = self.ds_tol)
        dsr = self.unitcell.ringds
        # npks
        npks = len(self.ds)
        self.ra = np.zeros(npks, np.int32)-1
        self.na = np.zeros(len(dsr), np.int32)
        print("Ring assignment array shape",self.ra.shape)
        tol = float(self.ds_tol)
        best = np.zeros(npks, np.float)+tol
        for j, dscalc in enumerate(dsr):
            dserr = abs(self.ds - dscalc)
            sel = dserr < best
            self.ra[sel]=j
            best[sel] = dserr[sel]
        # Report on assignments
        ds=np.array(self.ds)
        print("Ring     (  h,  k,  l) Mult  total indexed to_index  ubis  peaks_per_ubi   tth")
        minpks = 0
        # try reverse order instead
        for j in range(len(dsr))[::-1]:
            ind = np.compress( np.equal(self.ra,j), np.arange(self.ra.shape[0]) )
            self.na[j]=ind.shape[0]
            n_indexed  = np.sum(np.where( self.ga[ind] >  -1, 1, 0))
            n_to_index = np.sum(np.where( self.ga[ind] == -1, 1, 0))
            # diffs = abs(take(ds,ind) - dsr[j])
            h=self.unitcell.ringhkls[dsr[j]][0]
            Mult = len(self.unitcell.ringhkls[dsr[j]])
            try:
                expected_orients = int(180./self.omega_fullrange * self.na[j]/float(Mult))
                expected_npks = int(self.omega_fullrange/180. * Mult)
                minpks += expected_npks
            except:
                expected_orients = 'N/A'
                expected_npks = 'N/A'
            tth = 2*np.degrees(np.arcsin(dsr[j]*self.wavelength/2))
            print("Ring %-3d (%3d,%3d,%3d)  %3d  %5d   %5d    %5d %5s     %2s  %.2f"%(
                j,h[0],h[1],h[2],Mult,
                self.na[j],n_indexed,n_to_index,expected_orients,expected_npks,tth))
        if minpks > 0:
            print('\nmin_pks:  - Current  --> %3d'%(self.minpks))
            print('          - Expected --> %3d\n'%(minpks))

        # We will only attempt to index g-vectors which have been assigned
        # to hkl rings (this gives a speedup if there
        # are a lot of spare peaks
        ind = np.compress(np.greater(self.ra,-1), np.arange(self.ra.shape[0]))
        self.gvr = self.gv[ind]
        print("Using only those peaks which are assigned to rings for scoring trial matrices")
        print("Shape of scoring matrix",self.gvr.shape)
        self.gvflat=np.ascontiguousarray(self.gvr, np.float) # Makes it contiguous
        # in memory, hkl fast index

    def friedelpairs(self,filename):
        """
        Attempt to identify Freidel pairs

        Peaks must be assigned to the same powder ring
        Peaks will be the closest thing to being 180 degrees apart
        """
        out = open(filename,"w")
        dsr=self.unitcell.ringds
        nring = len(dsr)
        for j in range( nring ):
            ind = np.compress(np.equal(self.ra,j), np.arange(self.ra.shape[0]))
            # ind is the indices of the ring assigment array - eg which hkl is this gv
            #
            if len(ind)==0:
                continue
            thesepeaks = self.gv[ind]
            #
            h=self.unitcell.ringhkls[dsr[j]][0]
            #
            out.write("\n\n\n# h = %d \n"%(h[0]))
            out.write("# k = %d \n"%(h[1]))
            out.write("# l = %d \n"%(h[2]))
            out.write("# npks = %d \n"%(thesepeaks.shape[0]))
            out.write("# score eta1 omega1 tth1 gv1_x gv1_y gv1_z eta2 omega2 tth2 gv2_x gv2_y gv2_z\n")
            for k in range(thesepeaks.shape[0]):
                nearlyzero = thesepeaks + thesepeaks[k]
                mag = np.sum(nearlyzero*nearlyzero,1)
                best = np.argmin(mag)
                if best > k:
                    a = ind[k]
                    out.write("%f "%( np.sqrt(mag[best]) ) )
                    out.write("%f %f %f %f %f %f    "%(self.eta[a],self.omega[a],self.tth[a],self.gv[a][0],self.gv[a][1],self.gv[a][2]))
                    a = ind[best]
                    out.write("%f %f %f %f %f %f    "%(self.eta[a],self.omega[a],self.tth[a],self.gv[a][0],self.gv[a][1],self.gv[a][2]))
                    out.write("\n")

    def score_all_pairs(self):
        """
        Generate all the potential pairs of rings and go score them too
        """
        self.assigntorings()
        # Which rings have peaks assigned to them?
        rings = [r for r in set( self.ra ) if r >= 0 ]
        # What are the multiplicities of these rings? We will use low multiplicity first
        mults = {r:len(self.unitcell.ringhkls[self.unitcell.ringds[r]]) for r in rings}
        # How many peaks per ring? We will use the sparse rings first...
        # why? We assume these are the strongest peaks on a weak high angle ring
        # occupation = {r:self.na[r] for r in rings}
        pairs = [(int(mults[r1]*mults[r2]),int(self.na[r1]*self.na[r2]),r1,r2) for r1 in rings for r2 in rings]
        pairs.sort()
        self.tried = 0
        self.npairs = len(pairs)
        self.stop=False
        for mu, oc, r1, r2 in pairs:
            try:
                self.ring_1 = r1
                self.ring_2 = r2
                self.find()
                self.scorethem()
                self.tried += 1
            except KeyboardInterrupt:
                break
            if self.stop:
                break
            print("Tried r1=%d r2=%d attempt %d of %d, got %d grains"%(r1,r2,self.tried,len(pairs),len(self.ubis)))
        print("\nTested",self.tried,"pairs and found",len(self.ubis),"grains so far")

    def find(self):
        """
        Dig out the potential hits
        """
        # Optionally only used unindexed peaks here. Make this obligatory
        # Need indices of gvectors to test.
        # Bug out early when there are none
        iall = np.arange(self.gv.shape[0])
        i1 = np.compress(np.logical_and(np.equal(self.ra,self.ring_1),
                                      self.ga==-1  ) , iall).tolist()
        i2 = np.compress(np.logical_and(np.equal(self.ra,self.ring_2),
                                      self.ga==-1  ) , iall).tolist()
        if len(i1) == 0 or len(i2) == 0:
            logging.info("no peaks left for those rings")
            return
        # Which are the rings being used for indexing
        hkls1 = self.unitcell.ringhkls[self.unitcell.ringds[int(self.ring_1)]]
        hkls2 = self.unitcell.ringhkls[self.unitcell.ringds[int(self.ring_2)]]
        logging.info("hkls of rings being used for indexing")
        logging.info("Ring 1: %s",str(hkls1))
        logging.info("Ring 2: %s",str(hkls2))
        cosangles=[]
        for h1 in hkls1:
            for h2 in hkls2:
                ca=self.unitcell.anglehkls(h1,h2)
                cosangles.append(ca[1])
        cosangles.sort()
        coses=[]
        while len(cosangles)>0:
            a=cosangles.pop()
            if abs(a-1.)<1e-5 or abs(a+1.)<1e-5: # Throw out 180 degree angles
                continue
            if len(coses)==0:
                coses.append(a)
                continue
            if abs(coses[-1]-a) > 1e-5:
                coses.append(a)
        logging.info("Possible angles and cosines between peaks in rings:")
        for c in coses:
            logging.info("%.6f %.6f"%(math.acos(c)*180/math.pi,c))
        #
        #
        logging.info("Number of peaks in ring 1: %d",len(i1))
        logging.info("Number of peaks in ring 2: %d",len(i2))
        logging.info("Minimum number of peaks to identify a grain %d",self.minpks)
        # print self.gv.shape
        # ntry=0
        # nhits=0
        self.hits=[]
        if len(i1)==0 or len(i2)==0:
            # return without crashing please
            return
        tol=float(self.cosine_tol)
        # ng=0
        mp=np.sqrt(np.sum(self.gv*self.gv,1))
        # print mp.shape
        ps1 = np.take(self.gv,i1,0)
        mp1 = np.take(mp,i1,0)
        n1 = ps1.copy()
        ps2 = np.take(self.gv,i2,0)
        mp2 = np.take(mp,i2,0)
        n2 = ps2.copy()
        # print "mp1.shape",mp1.shape
        # print "n1[:,1].shape",n1[:,1].shape
        for i in range(3):
            n1[:,i]=n1[:,i]/mp1
            n2[:,i]=n2[:,i]/mp2
        cs = np.array(coses,'d')
        # found=0
        hits=[]
        start = time.time()
        mtol = -tol # Ugly interface - set cosine tolerance negative for all
                    # instead of best
        for i in range(len(i1)):
            costheta=np.dot(n2,n1[i])
            if tol > 0: # This is the original algorithm - the closest angle
                best,diff = cImageD11.closest(costheta,cs)
                if diff < tol:
                    hits.append( [ diff, i1[i], i2[best] ])
            else:
                for cval in cs:
                    # 1d   scalar  1d
                    diff = cval - costheta
                    candidates = np.compress( abs(diff) < mtol, i2 )
                    for c in candidates:
                        hits.append( [ 0.0, i1[i], c ] )
        logging.info("Number of trial orientations generated %d",len(hits))
        logging.info("Time taken %.6f /s",time.time()-start)
        self.hits=hits

    def histogram_drlv_fit(self,UBI=None,bins=None):
        """
        Generate a histogram of |drlv| for a ubi matrix
        For use in validation of grains
        """
        if UBI is None:
            ubilist = self.ubis
        else:
            ubilist = [UBI]
        if bins is None:
            start=0.25
            fac=2
            bins=[start]
            while start > 1e-5:
                start=start/fac
                bins.append(start)
            bins.append(-start)
            bins.reverse()
            bins=np.array(bins)
        hist = np.zeros((len(ubilist),bins.shape[0]-1),np.int)
        j=0
        for UBI in ubilist:
            drlv2 = calc_drlv2(UBI, self.gv)
            drlv = np.sort(np.sqrt(drlv2)) # always +ve
            if drlv[-1]>0.866:
                print("drlv of greater than 0.866!!!",drlv[-1])
            positions =  np.searchsorted(drlv,bins)
            hist[j,:] =  positions[1:]-positions[:-1]
            j=j+1
        self.bins=bins
        self.histogram=hist

    def scorethem(self, fitb4=False):
        """ decide which trials listed in hits to keep """
        start=time.time()
        ng=0
        tol=float(self.hkl_tol)
        gv=self.gvflat
        all=len(self.hits)
        logging.info("Scoring %d potential orientations",all)
        progress=0
        nuniq=0
        while len(self.hits) > 0 and ng < self.max_grains:
            diff,i,j = self.hits.pop()
            if self.ga[i]>-1 or self.ga[j]>-1 or i==j:
                # skip things which are already assigned or errors
                continue
            try:
                self.unitcell.orient(self.ring_1,
                                     self.gv[i,:],
                                     self.ring_2,
                                     self.gv[j,:],
                                     verbose=0,
                                     all=False)
            except:
                logging.error(str(i,j,self.ring_1,self.ring_2))
                logging.error(str(self.gv[i]))
                logging.error(str(self.gv[j]))
                logging.error("Failed to find orientation in unitcell.orient")
                raise
            if fitb4: # FIXME : this does not work
               self.unitcell.UBI = ubi_fit_2pks( self.unitell.UBI, self.gv[i,:], self.gv[j,:])
            npk = cImageD11.score(self.unitcell.UBI,gv,tol)
            if npk > self.minpks:
                self.unitcell.orient(self.ring_1, self.gv[i,:], self.ring_2, self.gv[j,:], verbose=0, all=True)
                if fitb4:
                   for k in range( len(self.unitell.UBIlist) ):
                        self.unitell.UBIlist[k] = ubi_fit_2pks( self.unitell.UBIlist[k],
                                                                self.gv[i,:], self.gv[j,:])
                npks=[cImageD11.score(UBItest,gv,tol) for UBItest in self.unitcell.UBIlist] 
                choice = np.argmax(npks)
                UBI = self.unitcell.UBIlist[choice]
                if npks[choice] < npk:
                    logging.error("Error in indexing: debug please!")
                    #import pdb; pdb.set_trace()
                npk = npks[choice]
                _ = cImageD11.score_and_refine( UBI, gv, tol )
                # See if we already have this grain...
                try:
                    ind=self.getind(UBI) # indices of peaks indexed
                    ga=self.ga[ind]  # previous grain assignments
                    uniqueness=np.sum(np.where(ga==-1,1,0))*1.0/ga.shape[0]
                    if uniqueness > self.uniqueness:
                        np.put(self.ga, ind, len(self.scores)+1)
                        self.ubis.append(UBI)
                        self.scores.append(npk)
                        logging.info("new grain %d pks, UBI %s",npk,str(UBI.ravel()))
                        ng=ng+1
                    else:
                        nuniq=nuniq+1
                except:
                    raise
                 
        logging.info("Number of orientations with more than %d peaks is %d",self.minpks,len(self.ubis))
        logging.info("Time taken %.3f/s",time.time()-start)
        if len(self.ubis)>0:
            bestfitting=np.argmax(self.scores)
            logging.info("UBI for best fitting\n%s",str(self.ubis[bestfitting]))
            logging.info("Unit cell: %s\n",str(ubitocellpars(self.ubis[bestfitting])))
            self.refine( self.ubis[bestfitting] )
            logging.info("Indexes %d peaks, with <drlv2>=%f",self.scorelastrefined,self.fitlastrefined)
            logging.info("That was the best thing I found so far")
            notaccountedfor =  ((self.ga < 0) & (self.ra >= 0)).sum()
            logging.info("Number of peaks assigned to rings but not indexed = %d", notaccountedfor)
        else:
            logging.info("Try again, either with larger tolerance or fewer minimum peaks")

    def fight_over_peaks(self):
        """
        Get the best ubis from those proposed
        Use all peaks (ring assigned or not)
        """
        self.drlv2 = np.zeros( self.gv.shape[0], np.float)+2
        labels = np.ones( self.gv.shape[0], np.int32)
        np.subtract(labels,2,labels)
        i = -1
        for ubi in self.ubis:
            i += 1
            try:
               npk = cImageD11.score_and_assign( ubi, self.gv, self.hkl_tol,
                                                 self.drlv2, labels, i)
            except:
               print(ubi.shape)
               print(self.gv.shape)
               print(self.hkl_tol)
               print(self.drlv2.shape)
               print(labels.shape)
               print("Error in fight_over_peaks",__file__)
               raise
        self.ga = labels
        # For each grain we want to know how many peaks it indexes
        # This is a histogram of labels
        bins =  np.arange(-0.5, len(self.ubis)-0.99)
        hst = myhistogram( labels, bins )
        self.gas = hst
        assert len(self.gas) == len(self.ubis)



    def saveindexing(self, filename, tol = None):
        """
        Save orientation matrices

        FIXME : refactor this into something to do
            grain by grain }
            peak by peak   }
        """
        f = open(filename,"w")
        i = 0
        from ImageD11 import transform

        self.gv = np.ascontiguousarray( self.gv )

        # grain assignment
        self.fight_over_peaks()

        # Printing per grain
        uinverses = []
        allind = np.array(list(range(len(self.ra))))
        tthcalc =np.zeros(len(self.ra),np.float)
        etacalc =np.zeros(len(self.ra),np.float)
        omegacalc = np.zeros(len(self.ra),np.float)
        i = -1

        
        for ubi in self.ubis:
            i += 1
            # Each ubi has peaks in self.ga
            uinverses.append( np.linalg.inv(ubi) )
            # self.ga was filled in during fight_over_peaks
            npk , mdrlv = cImageD11.refine_assigned(
                ubi.copy(),
                self.gv,
                self.ga,
                i)
            assert npk == self.gas[i]
            f.write("Grain: %d   Npeaks=%d   <drlv>=%f\n"%(
                i, self.gas[i], np.sqrt(mdrlv) ))
            f.write("UBI:\n"+str(ubi)+"\n")
            cellpars = ubitocellpars(ubi)
            f.write("Cell pars: ")
            for abc in cellpars[:3]: f.write("%10.6f "%(abc))
            for abc in cellpars[3:]: f.write("%10.3f "%(abc))
            f.write("\n")
            # Grainspotter U
            f.write("U:\n"+str( ubitoU(ubi) ) + "\n")
            f.write("B:\n"+str( ubitoB(ubi) ) + "\n")

            # Compute hkls
            h = np.dot( ubi, self.gv.T )
            hint = np.floor( h + 0.5 )
            gint= np.dot( uinverses[-1], hint )
            dr = h - hint

            f.write(
                "Peak   (  h       k       l      )   drlv             x       y ")
            if self.wavelength < 0:
                f.write("\n")
            else:
                f.write(
                    "   Omega_obs Omega_calc   Eta_obs Eta_calc   tth_obs tth_calc\n")

                tc, ec, oc =  transform.uncompute_g_vectors(
                    gint,
                    self.wavelength,
                    wedge = self.wedge)
                ind = np.compress( self.ga == i, allind)

            for j in ind:
                f.write("%-6d ( % 6.4f % 6.4f % 6.4f ) % 12.8f "%(j,h[0,j],h[1,j],h[2,j],np.sqrt(self.drlv2[j])) )
                f.write(" % 7.1f % 7.1f "%(self.xp[j],self.yp[j]) )
                if self.wavelength < 0:
                    f.write("\n")
                else:
                    # # # These should be equal to
                    to = math.asin( self.wavelength * self.ds[j]/2)*360/math.pi
                    # tth observed
                    eo = mod_360(self.eta[j], 0)
                    oo = self.omega[j]
                    tc1 = tc[j]
                    # Choose which is closest in eta/omega,
                    # there are two choices, {eta,omega}, {-eta,omega+180}
                    w = np.argmin( [ abs(ec[0][j] - eo) , abs(ec[1][j] - eo) ] )
                    ec1 = ec[w][j]
                    oc1 = oc[w][j]
                    # Now find best omega within 360 degree intervals
                    oc1 = mod_360(oc1, oo)
                    f.write("  % 9.4f % 9.4f     % 9.4f % 9.4f   % 9.4f % 9.4f"% (oo,oc1, eo,ec1 ,to,tc1))
                    etacalc[ j ] = ec1
                    omegacalc[ j ] = oc1
                    tthcalc[ j ] = tc1
                if self.ra[j] == -1 :
                    f.write(" *** was not assigned to ring\n")
                else:
                    f.write("\n")
            f.write("\n\n")
        # peaks assigned to rings
        in_rings = np.compress(np.greater(self.ra,-1),np.arange(self.gv.shape[0]))
        f.write("\n\nAnd now listing via peaks which were assigned to rings\n")
        nleft=0
        nfitted=0
        npk = 0
        for peak in in_rings:
            # Compute hkl for each grain
            h = self.gv[peak,:]
            f.write("\nPeak= %-5d Ring= %-5d gv=[ % -6.4f % -6.4f % -6.4f ]   omega= % 9.4f   eta= % 9.4f   tth= % 9.4f\n"%(peak,self.ra[peak],h[0],h[1],h[2],
                  self.omega[peak],self.eta[peak],self.tth[peak]))
            if self.ga[peak] != -1:
                m = self.ga[peak]
                hi = np.dot(self.ubis[m],h)
                hint = np.floor(hi+0.5).astype(np.int)
                gint = np.dot(uinverses[m], hint)
                f.write("Grain %-5d (%3d,%3d,%3d)"%( m,
                    hint[0],hint[1],hint[2]))
                f.write("  ( % -6.4f % -6.4f % -6.4f )  "%(hi[0],hi[1],hi[2]))
                # Now find best omega within 360 degree intervals
                f.write(" omega= % 9.4f   eta= %9.4f   tth= %9.4f\n"%(
                    omegacalc[peak],etacalc[peak],tthcalc[peak]) )
                npk=npk+1
            else:
                if len(self.ubis)>0:
                    f.write("Peak not assigned\n")
                    # , closest=[ % -6.4f % -6.4f % -6.4f ] for grain %d\n"%(hi[0],hi[1],hi[2],m))
                else:
                    f.write("Peak not assigned, no grains found\n")
                nleft=nleft+1

        f.write("\n\nTotal number of peaks was %d\n"%(self.gv.shape[0]))
        f.write("Peaks assigned to grains %d\n"%(npk))
        f.write("Peaks assigned to rings but remaining unindexed %d\n"%(nleft))

        f.write("Peaks not assigned to rings at all %d\n"%(np.sum(np.where(self.ra==-1,1,0))))
        f.close()




    def getind(self,UBI,tol=None):
        """
        Returns the indices of peaks in self.gv indexed by matrix UBI
        """
        if tol == None:
            tol = self.hkl_tol
        ng = len(self.gvflat)
        drlv2 = np.ones( ng , np.float)
        labels = np.full( ng , 0, np.int32)
        # we only use peaks assigned to rings for scoring here
        # already done in making gvflat in assigntorings
        try:
           npk = cImageD11.score_and_assign( UBI, self.gvflat, tol, drlv2, labels, 1 )
        except:
           print(self.gvflat.shape)
           print('ra',self.ra.shape)
           print(drlv2.shape)
           print(labels.shape)
           print("logic error in getind")
           raise
        ind = np.compress( labels , np.arange(ng,dtype=np.int) )
        return ind


    def score(self,UBI,tol=None):
        """
        Decide which are the best orientation matrices
        """
        if tol is None:
            return cImageD11.score(UBI,self.gvflat,self.hkl_tol)
        else:
            return cImageD11.score(UBI,self.gvflat,tol)

    def refine(self,UBI):
        """
        Refine an orientation matrix and rescore it.

        From Paciorek et al Acta A55 543 (1999)
        UB = R H-1
           where:
           R = sum_n r_n h_n^t
           H = sum_n h_n h_n^t
           r = g-vectors
           h = hkl indices
        """
#      print "Orientation and unit cell refinement of"
#      print "UBI\n",UBI
#      print "Scores before",self.score(UBI)
        # Need to find hkl indices for all of the peaks which are indexed
        drlv2=calc_drlv2(UBI,self.gv)
        h = np.dot(UBI, np.transpose(self.gv))
        hint = np.floor(h + 0.5).astype(np.int) # rounds down
        tol = float(self.hkl_tol)
        tol = tol*tol
        # Only use peaks which are assigned to rings for refinement
        ind = np.compress( np.logical_and(np.less(drlv2,tol),np.greater(self.ra,-1)) , np.arange(self.gv.shape[0]) )
        #scoreb4=ind.shape[0]
        contribs = drlv2[ind]
        if len(contribs)==0:
            raise Exception("No contributing reflections for"+str(UBI))
        #try:
        #    fitb4=sum(contribs)/contribs.shape[0]
        #except:
        #    print "No contributing reflections for\n",UBI
        #    raise
        #drlv2_old=drlv2
        R=np.zeros((3,3),np.float)
        H=np.zeros((3,3),np.float)
        for i in ind:
            r = self.gv[i,:]
            k = hint[:,i].astype(np.float)
#           print r,k
            R = R + np.outer(r,k)
            H = H + np.outer(k,k)
        try:
            HI=np.linalg.inv(H)
            UBoptimal=np.dot(R,HI)
            UBIo=np.linalg.inv(UBoptimal)
        except:
            # A singular matrix - this sucks.
            UBIo=UBI
        drlv2 = calc_drlv2(UBIo,self.gv)
        ind = np.compress( np.logical_and(np.less(drlv2,tol),np.greater(self.ra,-1)), np.arange(self.gv.shape[0]) )
        self.scorelastrefined=ind.shape[0]
        contribs = drlv2[ind]
        try:
            self.fitlastrefined=math.sqrt(np.sum(contribs)/contribs.shape[0])
        except:
            print("\n\n\n")
            print("No contributing reflections for\n",UBI)
            print("After refinement, it was OK before ???")
            print("\n\n\n")
            raise
#      for i in ind:
#         print "( %-6.4f %-6.4f %-6.4f ) %12.8f %12.8f"%(h[0,i],h[1,i],h[2,i],sqrt(drlv2[i]),sqrt(drlv2_old[i]))
#      print UBIo
#      print "Scores after", self.score(UBIo,self.hkl_tol)
#      print "diff\n",UBI-UBIo
#      print "Mean drlv now",sum(sqrt(drlv2))/drlv2.shape[0],
#      print "Mean drlv old",sum(sqrt(drlv2_old))/drlv2_old.shape[0]
        return UBIo

    def saveubis(self,filename):
        """
        Save the generated ubi matrices into a text file
        """
        write_ubi_file(filename, self.ubis)


    def coverage(self):
        """
        Compute the expected coverage of reciprocal space
        use the min/max obs values of xp/yp/omega to work out what was measured in the scan?
        No lambda or
        """
        pass


    def readgvfile(self,filename, quiet=False):
        f=open(filename,"r")
        # Lattice!!!
        self.unitcell = unitcell.cellfromstring(f.readline())
        while 1:
            line=f.readline()
            if line[0]=="#":
                if line.find("wavelength")>-1:
                    self.wavelength = float(line.split()[-1])
                    if not quiet:
                        print("Got wavelength from gv file of ",self.wavelength)
                    continue
                if line.find("wedge")>-1:
                    self.wedge = float(line.split()[-1])
                    if not quiet:
                        print("Got wedge from gv file of ",self.wedge)
                    continue
                if line.find("ds h k l")>-1:
                    continue   # reads up to comment line
                if line.find("omega")>-1 and line.find("xc")>-1:
                    break
        self.eta=[]   # Raw peak information
        self.omega=[]
        self.ds=[]
        self.xr=[]
        self.yr=[]
        self.zr=[]
        self.xp=[]
        self.yp=[]
        self.omega_ranges=[]
        self.omega_fullrange = 0
        for line in f.readlines():
            try:
                v=[float(x) for x in line.split()]
                if len(v) == 0:
                    # Skip the blank lines
                    continue
                if self.out_of_eta_range(v[6]):
                    continue
                self.xr.append(v[0])
                self.yr.append(v[1])
                self.zr.append(v[2])
                self.xp.append(v[3])
                self.yp.append(v[4])
                self.ds.append(v[5])
                self.eta.append(v[6])
                self.omega.append(v[7])
            except:
                print("LINE:",line)
                raise
#            raise "Problem interpreting the last thing I printed"
        f.close()
        self.ds = np.array( self.ds )
        self.omega_ranges.append(int(round(min(self.omega))))
        sorted_omega = sorted(self.omega)
        for i in range(len(sorted_omega)-1):
            if sorted_omega[i+1]-sorted_omega[i] > 10:
                self.omega_ranges.append(int(round(sorted_omega[i])))
                self.omega_ranges.append(int(round(sorted_omega[i+1])))
        self.omega_ranges.append(int(round(max(self.omega))))
        if not quiet:
            print('\nGot %d sets of omega values from gv file:'%(len(self.omega_ranges)/2))
        for i in range(len(self.omega_ranges)//2): # integer division intended here
            self.omega_fullrange += self.omega_ranges[2*i+1]-self.omega_ranges[2*i]
            if not quiet:
                print('- Range %1d from %4d to %4d ==> %3d degrees'%(i+1, \
                       self.omega_ranges[2*i], self.omega_ranges[2*i+1], \
                       self.omega_ranges[2*i+1]-self.omega_ranges[2*i]))

        if self.wavelength > 0:
            self.tth=np.arcsin(np.array(self.ds)*self.wavelength/2)*360/math.pi
        else:
            self.tth=np.zeros(len(self.ds))
        self.gv=np.transpose(np.array( [ self.xr , self.yr, self.zr ] ,np.float))
        self.allgv = self.gv.copy()
        self.ga=np.zeros(len(self.ds),np.int32)-1 # Grain assignments

        self.gvflat=np.ascontiguousarray(self.gv,np.float)
        # Makes it contiguous in memory, hkl fast index
        if not quiet:
            print("Read your gv file containing",self.gv.shape)



#!/usr/bin/env python

from __future__ import print_function

"""
Reciprocal space volume mapper
Transfers images into reciprocal space by pixel mapping
"""

import numpy, logging
from ImageD11 import parameters, transform, indexing, \
    cImageD11, blobcorrector, rsv, ImageD11options


class rsv_mapper(object):
    """
    Maps images into a reciprocal space volume

    Basic idea is for a fixed detector, scattering vectors in lab
    frame are fixed. When rotating sample, just rotate these into
    crystal frame and use as array indices into volume
    """
    def __init__(self, dims, pars, ubi, 
                 splinefile = None,
                 np=16, 
                 border = 10, 
                 omegarange = list(range(360)),
                 maxpix = None,
                 mask = None):
        """
        Create a new mapper intance. It will transform images into 
        reciprocal space (has its own rsv object holding the space)
        dims - image dimensions
        par  - ImageD11 parameter filename for experiment
        ubi  - Orientation matrix (ImageD11 style)
        np   - Number of pixels per hkl index [16]
        border - amount to add around edge of images [10]
        omegarange - omega values to be mapped (0->360)
        maxpix - value for saturated pixels to be ignored
        mask - fit2d style mask for removing bad pixels / border
        """
        if len(dims)!=2: raise Exception("For 2D dims!")
        self.dims = dims
        print(dims)
        # Experiment parameters
        if not isinstance( pars, parameters.parameters):
            raise Exception("Pars should be an ImageD11 parameters object")
        for key in ["distance", "wavelength"]: #etc
            assert key in pars.parameters
        self.pars = pars

        # Orientation matrix
        self.ubi = ubi

        # Saturation
        self.maxpix = maxpix

        # Mask
        self.mask = mask
        if self.mask is not None:
            assert self.mask.shape == self.dims, "Mask dimensions mush match image"
        
        # spatial
        if splinefile is None:
            self.spatial = blobcorrector.perfect()
        else:
            self.spatial = blobcorrector.correctorclass(splinefile)

        # npixels
        self.np  = np

        self.uspace = np*ubi

        self.find_vol( border = border, omegarange = omegarange )

        self.rsv.metadata['ubi'] = ubi
        self.rsv.metadata['uspace'] = self.uspace
        # Make and cache the k vectors
        self.make_k_vecs()
        
        
    def find_vol( self, border , omegarange ):
        """
        find limiting volume
        The four image corners over 360 degrees

        np is the number of pixels per hkl

        returns (RSV, NR)
           RSV min/max in reciprocal space (np*ubi).gv
           NR number of points == 1+RSV[i][1]-RSV[i][0]
        """
        # Note that ImageD11 peaks are [slow, fast]
        #  1   2   3
        #  4   5   6    
        #  7   8   9
        p1 = [ -border, self.dims[0]/2, self.dims[0]+border ,
                self.dims[0]/2, self.dims[0]+border ,-border,
                self.dims[0]+border, -border, self.dims[0]/2 ]
        p2 = [ -border, self.dims[1]/2, self.dims[1]+border ,
                -border, self.dims[1]/2, self.dims[1]+border ,
                -border, self.dims[1]/2, self.dims[1]+border ]
        for i in range(9):
            p1[i], p2[i] = self.spatial.correct( p1[i], p2[i] )
        peaks   = [p1*len(omegarange), p2*len(omegarange)]
        om = numpy.array(list(omegarange)*9, numpy.float32)
        tth, eta = transform.compute_tth_eta( peaks,
                                              **self.pars.get_parameters() )
        #print "tth",tth.min(),tth.max()
        #print "eta",eta.min(),eta.max()
        assert om.shape == tth.shape
        gv = transform.compute_g_vectors( tth, eta, om,
                                          self.pars.get('wavelength'),
                                          float(self.pars.get('wedge')),
                                          float(self.pars.get('chi')))
                                          
    
        # Rotate g-vectors into target volume
        hkls = numpy.dot( self.uspace, gv )

        # print "Ranges for RSV"
        bounds=numpy.zeros((3,2))
        npv =numpy.zeros(3)
        for i in range(3):
            bounds[i][0] = numpy.floor(hkls[i].min())
            bounds[i][1] = numpy.ceil(hkls[i].max())
            npv[i] = (bounds[i][1]-bounds[i][0]) + 1
        self.bounds = bounds
        self.rsv = rsv.rsv( npv , bounds=bounds, np=self.np )
        # Cross your fingers and....
        self.rsv.allocate_vol()

                            

    def make_k_vecs( self ):
        """
        Generate the k vectors from the experiment parameters
        given in constructor
        """
        xim, yim = self.spatial.make_pixel_lut( self.dims )
        peaks = [ numpy.ravel(xim), numpy.ravel(yim) ]
        # First, x, is the slow pixel direction, should not change
        # when raveled
        #for i in range(10):
        #    print "slow, fast"
        #    print peaks[0][i],peaks[1][i]
        assert abs(peaks[0][10]-peaks[0][0]) < 3
        # Second, y, is the fast, should change by ~ 1 pixel per pixel
        assert abs(peaks[1][10]-peaks[1][0]-10) < 3
        tth, eta = transform.compute_tth_eta( peaks, 
                                              **self.pars.get_parameters() )
        self.k = transform.compute_k_vectors(tth, eta,
                                             self.pars.get('wavelength'))
                                             
        # FIXME
        # This should be something like domega/dk where
        #      dk [ k(omega=0) - k(omega=1) ]
        
        self.lorfac = numpy.ones( self.dims[0]*self.dims[1],
                                  numpy.float32)
                        


    def add_image( self, om, data ):
        """
        RSV = bounds of reciprocal space vol
        NR = dims of RSV
        k = scattering vector for image at om == 0
        data = 2D image (dims == k.dims)
        SIG = signal
        MON = monitor
        """
        dat = numpy.ravel(data).astype(numpy.float32)
        assert len(dat) == len(self.k[0]), "dimensioning issue"
    
        # hkl = ubi.( gtok. k )
        gvm = transform.compute_g_from_k(numpy.eye(3),
                                         # this transform module sucks
                                         om*float(self.pars.get('omegasign')),
                                         wedge = float(self.pars.get('wedge')),
                                         chi = float(self.pars.get('chi')))
        tmat = numpy.dot( self.uspace , gvm )
        hkls = numpy.dot( tmat , self.k )

        # Find a way to test if we are doing the transform OK
    
        # Bounds checks
#        for i in range(3):
#            assert hkls[i].min() > self.bounds[i][0], \
#                "%d %s %s"%(i, str(hkls[i].min()),str( self.bounds[i][0]))
#            assert hkls[i].max() < self.bounds[i][1], \
#                "%d %s %s"%(i, str(hkls[i].max()),str( self.bounds[i][1]))

        NR = self.rsv.NR        
        # hkls[0] is the slowest index. integer steps of NR[1]*NR[2]
        ind = numpy.floor(hkls[0]+0.5-self.bounds[0][0]).astype(numpy.intp)
        numpy.multiply( ind, NR[1]*NR[2] , ind )
        assert ind.dtype == numpy.intp
        # hkls[1] is faster. Steps by NR[2] only
        numpy.add( ind, NR[2]*numpy.floor(
                hkls[1] + 0.5 - self.bounds[1][0]).astype(numpy.intp),
                   ind )
        numpy.add( ind, numpy.floor(
                hkls[2] + 0.5 - self.bounds[2][0]).astype(numpy.intp),
                   ind )
        #
        #
        #
        if self.maxpix is not None:
            msk =  numpy.where( dat > self.maxpix, 0, 1).astype(numpy.uint8)
        else:
            msk = None

        if self.mask is not None:
            # This excludes saturated pixels
            if msk is None:
                msk = self.mask
            else:
                numpy.multiply(msk, numpy.ravel(self.mask), msk)
                
        # cases:
        #    maxpix only     == msk
        #    mask only       == msk
        #    maxpix and mask == msk
        #    neither      


        if msk is not None:
            numpy.multiply(dat, msk, dat)
            cImageD11.put_incr( self.rsv.SIG,
                              ind,
                              dat )
            cImageD11.put_incr( self.rsv.MON,
                              ind,
                              self.lorfac * msk)
        else:
            cImageD11.put_incr( self.rsv.SIG,
                              ind,
                              dat )
            cImageD11.put_incr( self.rsv.MON,
                              ind,
                              self.lorfac)
        return

    def writevol(self, filename):
        """
        Save the volume in a hdf file
        """
        rsv.writevol( self.rsv, filename )



from ImageD11 import ImageD11_file_series

def get_options(parser):
    """
    Command line interface for making a mapping
    Add our options to a parser object
    """
    parser = ImageD11_file_series.get_options( parser )

    parser.add_argument("-p", "--pars", action="store",
                      dest = "pars", default = None,
                      type=ImageD11options.ParameterFileType(mode='r'),
                      help = "ImageD11 parameter file for experiment")
    
    parser.add_argument("-o", "--output", action="store",
                      dest = "output", default = None,
                      type=ImageD11options.HdfFileType(mode='r'),
                      help = "Name of hdf5 output file")

    parser.add_argument("-s", "--splinefile", action="store", 
                      dest = "spline", default = None,
                      type=ImageD11options.SplineFileType(mode='r'),
                      help = "Name of fit2d spline file for spatial dist")

    parser.add_argument("-u", "--ubifile", action="store", 
                      dest = "ubifile", default = None,
                      type = ImageD11options.UbiFileType(mode='r'),
                      help = "Name of ubi file (first matrix is used)")

    parser.add_argument("-x", "--npixels", action="store", type=int,
                      dest = "npixels", default = 16,
      help = "Number of pixels in reciprocal space map per integer hkl [16]")

    parser.add_argument("-i", "--images", action="store", type=int,
                      dest = "images", default = None,
                      help = "Number of images to process [all]")

    parser.add_argument("-b", "--border", action="store", type=int,
                       dest = "border", default = 10,
                       help = "Border around images to allocate space, px [10]")
    parser.add_argument("-t", "--saturation", action="store", type=float,
                      dest = "maxpix", default = None,
                      help = "Saturation value for excluding pixels")


    #parser.add_argument("-t", "--testcolfile", action="store", type="string",
    #                  dest = "testcolfile", default=None,
    #                  help = "A columnfile to test geometry")

    parser.add_argument("-c", "--subslice", action="store", type=int,
                      dest = "subslice", default=1,
                      help = "Number of omega subslices to repeat images")

    parser.add_argument("--maskfilename", action="store", type=str,
                      dest = "maskfilename", default=None,
                      help = "Mask image (fit2d style)" )
    
    return parser
                      
    
                    

def main():
    """
    A user interface
    """
    import sys, time, os, logging
    start = time.time()

    root = logging.getLogger('')
    root.setLevel(logging.WARNING)

    try:
        from argparse import ArgumentParser
        parser = ArgumentParser()
        parser = get_options( parser )
        options, args = parser.parse_known_args()
    except SystemExit:
        raise
    except:
        parser.print_help()
        print("\nProblem with your options:")
        raise
    
    if options.output is None:
        print("You must supply an output file (-o vol.h5)")
        sys.exit()

    if os.path.exists( options.output ):
        print("I would overwrite your output file",options.output)
        print("If you really want that then delete it first and re-run")
        sys.exit()
    
    try:
        if options.pars is None:
            print("You must supply a parameter file, -p file.pars")
            sys.exit()
        pars = parameters.parameters()
        pars.loadparameters(options.pars)
        print("Got parameters from",options.pars)
        pd = pars.get_parameters()
        names = list(pd.keys())
        names.sort()
        for name in names:
            print("%30s   %s"%(name, pd[name]))
    except:
        print("Problem with parameters:",options.pars)
        raise

    try:
        if options.ubifile is None:
            print("You must supply an input ubifile")
        ubi = indexing.readubis(options.ubifile)[0]
        print("UBI:\n",ubi)
        print("Cell parameters:")
        print("%.5f %.5f %.5f %.4f %.4f %.4f" % \
              indexing.ubitocellpars(ubi))
    except:
        print("Problem with ubi file:",options.ubifile)
        raise

    if options.maskfilename is not None:
        from fabio.openimage import openimage
        try:
            mask = ( openimage( options.maskfilename ).data == 0 )
        except:
            print("Problem with your mask image",options.maskfilename)
            raise
        print("Using a mask from",options.maskfilename)
        print("percent of image used %.3f"%(100.0*mask.sum()/mask.shape[0]/mask.shape[1]))
    else:
        mask = None
        
    
    first_image = True
    nimage = 0

    imagefiles = ImageD11_file_series.get_series_from_options( options, args )

    print("Subslicing by",options.subslice)

    try:
        for fim in imagefiles:
            
            if first_image: # allocate volume, compute k etc
                
                first_image = False
                
                mapper = rsv_mapper( fim.data.shape, 
                                     pars , ubi, 
                                     options.spline,
                                     np = options.npixels,
                                     border = options.border,
                                     maxpix = options.maxpix,
                                     mask = mask
                                     # FIXME omegarange
                                     )
                                
                logging.info( "Setting up time %.4f s"%(time.time()-start))

                
            ltp = time.time()
            om = float(fim.header['Omega'])
            oms = float(fim.header['OmegaStep'])
            for i in range(options.subslice):
                print(".", end=' ')
                omv = om + i*oms/options.subslice
                # ==1 : 0*s/1
                # ==2 : 0*s/2 , 1*s/2
                # ==3 : 0*s/3 , 1*s/3, 2*s/3 etc
                mapper.add_image( omv, fim.data )
                    
                    
            nimage = nimage + 1
            
            print("  %d %.3f %.4f s, %.4f s"%(nimage, om, 
                                            time.time()-ltp,
                                            time.time()-start))
            if options.images is not None:
                if nimage >= options.images:
                    break

    except KeyboardInterrupt:
        print("\nCaught a control-c")
        if nimage > 0:
            print("Problem, trying to save volume so far to:",options.output)
            mapper.writevol( options.output )
            print("Saved what I had")
        sys.exit()
    except:
        print("\nAn error occured")
        if nimage > 0:
            print("Problem, trying to save volume so far to:",options.output)
            mapper.writevol( options.output )
            print("Saved what I had")
        raise

    if nimage > 0:
        mapper.writevol( options.output )
        
            


        
if __name__=="__main__":
    main()



#!/usr/bin/env python

from __future__ import print_function

"""
from example by Tarn Weisner Burton <twburton@users.sourceforge.net> in pyopengl
"""

__author__ = 'Jon Wright <jpwright@users.sourceforge.net> from example by Tarn Weisner Burton <twburton@users.sourceforge.net>'

import numpy
import sys
import os
from pyopengltk import Opengl
import OpenGL.GL as GL
import OpenGL.GLU as GLU
if sys.version_info[0] < 3:
    import Tkinter as Tk
else:
    import tkinter as Tk


class myOpengl(Opengl):


    # Make a parallel projection
    # mostly copied from Tk.Opengl class with small mods
    def tkRedraw(self, *dummy):
        """Cause the opengl widget to redraw itself."""
        if not self.initialised: 
            return
        self.activate()
        #print self.distance
        GL.glPushMatrix()			# Protect our matrix
        self.update_idletasks()
        self.activate()
        w = self.winfo_width()
        h = self.winfo_height()
        GL.glViewport(0, 0, w, h)
        # Clear the background and depth buffer.
        GL.glClearColor(self.r_back, self.g_back, self.b_back, 0.)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        r = 1.*w/h
        GL.glOrtho( -self.distance*r, self.distance*r, -self.distance, self.distance, 
                     -self.distance*3, self.distance*3)
#        GLU.gluPerspective(self.fovy, float(w)/float(h), self.near, self.far)
        GL.glMatrixMode(GL.GL_MODELVIEW)
        self.redraw(self)
        GL.glFlush()				# Tidy up
        GL.glPopMatrix()			# Restore the matrix
#        self.tk.call(self._w, 'swapbuffers')
        self.tkSwapBuffers()



class plot3d(Tk.Toplevel):
    def __init__(self,parent,data=None,lines=None,
                 ubis=None,image=None,pars=None,spline=None):
        """
        Data would be your observed g-vectors. Lines will
        be a computed lattice
        """
        Tk.Toplevel.__init__(self,parent)
        self.parent=parent
        if data is not None:
            xyz=data.copy()
        else:
            xyz=numpy.array([0,0,0])
        self.ps=Tk.StringVar()
        self.ps.set('1.')
        self.pointsize=1.
        self.npeaks=xyz.shape[0]

        self.o = myOpengl(self, width = 400, height = 400)
        self.o.redraw = self.redraw
        self.o.autospin_allowed = 1
        self.o.fovy=5
        self.o.near=1e6
        self.o.far=1e-6
        import math
        self.o.distance=3.
#numpy.maximum.reduce(numpy.ravel(xyz))*4 / \
#            math.tan(self.o.fovy*math.pi/180)
        print(type(xyz),xyz.dtype.char,xyz.shape)
        self.xyz=xyz
        f=Tk.Frame(self)
        Tk.Button(f,text="Help",command=self.o.help).pack(side=Tk.LEFT)
        Tk.Button(f,text="Reset",command=self.o.reset).pack(side=Tk.LEFT)
        Tk.Button(f,text="Pointsize",command=self.setps).pack(side=Tk.LEFT)
        Tk.Entry(f,textvariable=self.ps).pack(side=Tk.LEFT)
        Tk.Button(f,text="Quit",command=self.goaway).pack(side=Tk.RIGHT)
        self.dataoff=0
        self.o.pack(side = 'top', expand = 1, fill = 'both')
        f.pack(side=Tk.BOTTOM,expand=Tk.NO,fill=Tk.X)
        Tk.Label(self,text="Red=[1,0,0] Green=[0,1,0] Blue=[0,0,1]").pack(
            side=Tk.BOTTOM,expand=Tk.NO,fill=Tk.X)
        self.ubis=ubis
        self.color=numpy.ones((xyz.shape[0],3),numpy.float)
        print(self.color.shape)
        self.tex=False
        if ubis is not None:
           self.ubis = self.readubis(ubis)
           self.scorecolor(0)
        if pars is not None:
           self.tex=True
           self.readspline(spline)
           self.readprms(pars)
           self.readimage(image)
        self.after(100, self.changedata)

    def readspline(self,spline):
        from ImageD11 import blobcorrector
        self.corrector = blobcorrector.correctorclass(spline)

    def readubis(self,ubis):
        from ImageD11 import indexing
        return indexing.readubis(ubis)

    def readprms(self,prms):
        from ImageD11 import parameters
        o = parameters.parameters()
        o.loadparameters(prms)
        self.pars=o.get_parameters()

    def readimage(self,image):
        from ImageD11 import transform
        from fabio import openimage
        self.imageobj=openimage.openimage(image)
        # map from 2048x2048 to 1024x1024
        d = self.imageobj.data.astype(numpy.float32)
        mi= d.mean() - d.std()*2
        mx= d.mean() * d.std()*2
        shape=self.imageobj.data.shape
        d=numpy.reshape(numpy.clip(self.imageobj.data,mi,mx),shape) # makes a clipped copy
        d=(255.*(d-mi)/(mx-mi)) # scale intensity
        print(d.min(),d.max(),d.mean())
        self.image=numpy.zeros((1024,1024),numpy.uint8)
        if d.shape==(2048,2048):
            # rebin 2x2
            im=(d[::2,::2]+d[::2,1::2]+d[1::2,::2]+d[1::2,1::2])/4
            self.image=(255-im).astype(numpy.uint8).tostring()
        self.imageWidth=1024
        self.imageHeight=1024
        # make a 2D array of x,y
        p=[]
        pk=[]
        step = 64
        r=[ [ 0,0 ], [0,step], [step,step], [step,0] ]
        for i in range(0,1024,step):
            for j in range(0,1024,step):
                # i,j 1024x1024 texture coords
                # x,y spatially corrected
                for v in r:
                    pk.append([i+v[0],j+v[1]])
                    x,y = self.corrector.correct((i+v[0])*2 , (j+v[1])*2) # corrected
                    p.append([x,y])
        p=numpy.array(p).T 
        pk=numpy.array(pk).T
        omega=float(self.imageobj.header['Omega'])
        self.pars['distance']=float(self.pars['distance'])*1000
        tth,eta=transform.compute_tth_eta(p,**self.pars)
        gve = transform.compute_g_vectors(tth,eta,omega*self.pars['omegasign'],self.pars['wavelength'])
        self.pts = []
        print("Setting up image mapping",p.shape,gve.shape)
        for i in range(pk.shape[1]):
            self.pts.append([pk[1,i]/1024.,pk[0,i]/1024.,gve[0,i],gve[1,i],gve[2,i]])
        #for p in self.pts:
        #    print p
        self.setupTexture()

    def setupTexture(self):
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glTexImage2D(GL.GL_TEXTURE_2D,#target
                    0,#level
                    3,#internalformat
                    self.imageWidth, self.imageHeight,
                    0,#border
                    GL.GL_LUMINANCE,#format
                    GL.GL_UNSIGNED_BYTE,# type
                    self.image)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_DECAL)
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glEnable(GL.GL_NORMALIZE)
        GL.glShadeModel(GL.GL_FLAT)



    def scorecolor(self,i=0):
        cc = [ [ 1,0,0] , [0,1,0] , [0,0,1], [1,1,0], [1,0,1], [0,1,1],
                [ 0.5,0,0] , [0,0.5,0] , [0,0,0.5], [0.5,0.5,0], [0.5,0,0.5],
                [0,0.5,0.5]]
        if self.ubis is not None:
            from ImageD11 import indexing
            for u,i in zip(self.ubis,list(range(len(self.ubis)))):
                scores=indexing.calc_drlv2(u,self.xyz)
                print(self.xyz.shape,scores.shape)
                ind = numpy.compress( numpy.less(scores,0.02) , 
                                      numpy.arange(self.xyz.shape[0]) )
                print("Grain",i,scores.shape,ind.shape)
                for j in range(3):
                    c=numpy.ones(self.color.shape[0])
                    numpy.put(c,ind,cc[i%len(cc)][j])
                    self.color[:,j]*=c

    def go(self):
        """
        Allow the toplevel to return a handle for changing data
        """
        self.o.mainloop()

    def goaway(self):
        print("Called goaway")
        self.o.destroy()
        self.destroy()
        if self.parent is None: sys.exit()
        print("Ought to be gone now...")

    def changedata(self,xyz=None):
        if xyz is not None:
            self.xyz=xyz.copy()
            self.npeaks=xyz.shape[0]
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        GL.glDisableClientState(GL.GL_COLOR_ARRAY)
        GL.glVertexPointer( 3, GL.GL_FLOAT, 0, self.xyz.astype(numpy.float32).tostring() )
        GL.glColorPointer( 3,  GL.GL_FLOAT, 0, self.color.astype(numpy.float32).tostring() )
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glEnableClientState(GL.GL_COLOR_ARRAY)
        self.o.tkRedraw()

    def setps(self):
        self.pointsize=float(self.ps.get())
        self.o.tkRedraw()



    def redraw(self,o):
        
        GL.glDisable(GL.GL_LIGHTING)
        GL.glClearColor(0., 0., 0., 0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        GL.glColor3f(1.0, 1.0, 1.0) # white
        GL.glPointSize(self.pointsize)
        GL.glDrawArrays(GL.GL_POINTS, 0, self.npeaks )

        if self.ubis is not None and len(self.ubis)==1:
            hkl = numpy.dot(numpy.linalg.inv(self.ubis[0]), 
                            numpy.identity(3,numpy.float)).T
            # print hkl
        else:
            hkl = numpy.identity(3,numpy.float)
            # print hkl
                            
        GL.glBegin(GL.GL_LINE_LOOP)
        GL.glColor3f(1.0, 0.0, 0.0) # red
        GL.glVertex3f(0.,0.,0.)
        GL.glVertex3f(hkl[0][0],hkl[0][1],hkl[0][2])
        GL.glEnd()

        GL.glBegin(GL.GL_LINE_LOOP)
        GL.glColor3f(0.0, 1.0, 0.0) # green
        GL.glVertex3f(0.,0.,0.)
        GL.glVertex3f(hkl[1][0],hkl[1][1],hkl[1][2])
        GL.glEnd()

        GL.glBegin(GL.GL_LINE_LOOP)
        GL.glColor3f(0.0, 0.0, 1.0) # blue
        GL.glVertex3f(0.,0.,0.)
        GL.glVertex3f(hkl[2][0],hkl[2][1],hkl[2][2])
        GL.glEnd()

        if self.tex:
#            print "drawing images"
            GL.glEnable(GL.GL_TEXTURE_2D)
            GL.glColor4f(.0, 1.0, .0, 1.0) # red        
            GL.glBegin(GL.GL_QUADS)
            # generate a grid of squares to map the texture in 3D
            # opengl has better "map" methods to do this
            for i,j,g1,g2,g3 in self.pts:
#                print i,j,g1,g2,g3
                GL.glTexCoord2f(i,j)
                GL.glVertex3f(g1, g2, g3) 
            GL.glEnd()
#            GL.glDisable(GL.GL_TEXTURE_2D)
            
        GL.glFlush()
        GL.glEnable(GL.GL_LIGHTING)







if __name__=="__main__":

    try:
        lines=open(sys.argv[1],"r").readlines()
    except:
        print("Usage %s gvector_file [ubifile] [image parfile]"%(sys.argv[0]))
        raise
        # sys.exit()
   
    on=0
    xyz=[]
    for line in lines:
        if on==1:
            try:
                vals=[float(x) for x in line.split()]
                xyz.append( [ vals[0],vals[1],vals[2] ])
            except:
                pass
        if line.find("xr yr zr")>0 or line.find("gx  gy  gz  ")>0:
            on=1
   
    npeaks = len(xyz)
    xyz=numpy.array(xyz)
    if len(sys.argv)==3:
       o=plot3d(None,data=xyz,ubis=sys.argv[2])
    elif len(sys.argv)==6:
       o=plot3d(None,data=xyz,ubis=sys.argv[2],image=sys.argv[3],pars=sys.argv[4],spline=sys.argv[5])
    else:
       o=plot3d(None,data=xyz,ubis=None)
    def runit():
        o.changedata(o.xyz)
    o.after(100, runit )
    o.mainloop()




from __future__ import print_function


#
# Look up where this comes from.
# It was copied from somewhere and modified slightly


try:
    import Tkinter as Tk
except:
    import tkinter as Tk



class listdialog(Tk.Toplevel):
    """
    Dialog box for setting detector parameters
    Takes a list of strings and numbers
    """
    def __init__(self, parent, title = None, items=None, logic = None):
        Tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        if title:
            self.title(title)
        self.logic = logic
        self.logicvars = {}
        self.parent = parent
        self.result = items
        body = Tk.Frame(self)
        self.initial_focus = self.body(body,items,logic)
        body.pack(padx=5, pady=5)
        self.buttonbox()
        self.grab_set()
        if not self.initial_focus:
            self.initial_focus = self
        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                  parent.winfo_rooty()+50))
        self.initial_focus.focus_set()
        self.wait_window(self)
        
    def body(self, master, items, logic=None):
        # create dialog body.  return widget that should have
        # initial focus.  this method should be overridden
        self.e=[]
        if items is not None:
            i=0
            keys=list(items.keys())
            keys.sort()
            self.keys=keys
            for key in keys:
                Tk.Label(master,text=key).grid(row=i)
                el=Tk.Entry(master)
                el.insert(Tk.END,items[key])
                el.grid(row=i,column=1)
                self.e.append(el)
                if logic is not None and key in logic:
                    val = logic[key]
                    self.logicvars[key] = Tk.IntVar()
                    self.logicvars[key].set(val)
                    b=Tk.Checkbutton(master,text="Vary?",
                                     variable=self.logicvars[key])
                    b.grid(row=i,column=2)
                i=i+1
            return self.e[0]

    def buttonbox(self):
        # add standard button box. override if you don't want the
        # standard buttons
        box = Tk.Frame(self)
        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)
        box.pack()
    #
    # standard button semantics
    def ok(self, event=None):
        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return
        self.withdraw()
        self.update_idletasks()
        self.apply()
        self.cancel()

    def cancel(self, event=None):
        # put focus back to the parent window
        self.parent.focus_set()
        self.destroy()
    #
    # command hooks
    def validate(self):
        return 1 # override

    def apply(self):
        retdict={}
        i=0
        self.fv = {}
        for item in self.e:
            k = self.keys[i]
            retdict[k]=item.get()
            if self.logic is not None and k in self.logic:
                self.fv[k]=self.logicvars[k].get()
            i=i+1
        self.result=retdict
        print(self.result)


class columnchooser(listdialog):
    """
    Dialog box for setting detector parameters
    Takes a list of strings and numbers
    """
    def __init__(self, parent, items, title="Choose two columns"):
        Tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        if title:
            self.title(title)
        body = Tk.Frame(self)
        listbox1 = Tk.Listbox(body)
        listbox2 = Tk.Listbox(body)
        for i in items:
            listbox1.insert(Tk.END,i)
            listbox2.insert(Tk.END,i)
        body.pack(padx=5, pady=5)
        listbox1.pack()
        listbox2.pack()
        self.initial_focus=body
        self.buttonbox()
        self.grab_set()
        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                  parent.winfo_rooty()+50))
        self.initial_focus.focus_set()
        self.wait_window(self)
        
